import sys
from _typeshed import ReadableBuffer
from typing import Callable

from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurve, EllipticCurvePrivateKey, EllipticCurvePublicKey
from paramiko.message import Message
from paramiko.transport import Transport

if sys.version_info >= (3, 0):
    from hashlib import _Hash
else:
    from hashlib import _hash as _Hash

c_MSG_KEXECDH_INIT: bytes
c_MSG_KEXECDH_REPLY: bytes

class KexNistp256:
    name: str
    hash_algo: Callable[[ReadableBuffer], _Hash]
    curve: EllipticCurve
    transport: Transport
    P: int | EllipticCurvePrivateKey
    Q_C: EllipticCurvePublicKey | None
    Q_S: EllipticCurvePublicKey | None
    def __init__(self, transport: Transport) -> None: ...
    def start_kex(self) -> None: ...
    def parse_next(self, ptype: int, m: Message) -> None: ...

class KexNistp384(KexNistp256):
    name: str
    hash_algo: Callable[[ReadableBuffer], _Hash]
    curve: EllipticCurve

class KexNistp521(KexNistp256):
    name: str
    hash_algo: Callable[[ReadableBuffer], _Hash]
    curve: EllipticCurve



from typing import Any

class OCSPVerifier:
    SOCK: Any
    HOST: Any
    PORT: Any
    CA_CERTS: Any
    def __init__(self, sock, host, port, ca_certs: Any | None = ...) -> None: ...
    def components_from_socket(self): ...
    def components_from_direct_connection(self): ...
    def build_certificate_url(self, server, cert, issuer_cert): ...
    def check_certificate(self, server, cert, issuer_url): ...
    def is_valid(self): ...



"""Printer to output pytd trees in pyi format."""

import collections
import copy
import logging
import re

from pytype import utils
from pytype.pytd import base_visitor
from pytype.pytd import pep484
from pytype.pytd import pytd
from pytype.pytd.parse import parser_constants


class PrintVisitor(base_visitor.Visitor):
  """Visitor for converting ASTs back to pytd source code."""
  visits_all_node_types = True
  unchecked_node_names = base_visitor.ALL_NODE_NAMES

  INDENT = " " * 4
  _RESERVED = frozenset(parser_constants.RESERVED +
                        parser_constants.RESERVED_PYTHON)

  def __init__(self, multiline_args=False):
    super().__init__()
    self.class_names = []  # allow nested classes
    self.imports = collections.defaultdict(set)
    self.in_alias = False
    self.in_parameter = False
    self.in_literal = False
    self.in_constant = False
    self.in_signature = False
    self.multiline_args = multiline_args

    self._unit = None
    self._local_names = {}
    self._class_members = set()
    self._typing_import_counts = collections.defaultdict(int)
    self._module_aliases = {}

  def Print(self, node):
    return node.Visit(copy.deepcopy(self))

  def _IsEmptyTuple(self, t):
    """Check if it is an empty tuple."""
    assert isinstance(t, pytd.GenericType)
    return isinstance(t, pytd.TupleType) and not t.parameters

  def _NeedsTupleEllipsis(self, t):
    """Do we need to use Tuple[x, ...] instead of Tuple[x]?"""
    assert isinstance(t, pytd.GenericType)
    if isinstance(t, pytd.TupleType):
      return False  # TupleType is always heterogeneous.
    return t.base_type == "tuple"

  def _NeedsCallableEllipsis(self, t):
    """Check if it is typing.Callable type."""
    assert isinstance(t, pytd.GenericType)
    return t.name == "typing.Callable"

  def _RequireImport(self, module, name=None):
    """Register that we're using name from module.

    Args:
      module: string identifier.
      name: if None, means we want 'import module'. Otherwise string identifier
       that we want to import.
    """
    self.imports[module].add(name)

  def _GenerateImportStrings(self):
    """Generate import statements needed by the nodes we've visited so far.

    Returns:
      List of strings.
    """
    ret = []
    for module in sorted(self.imports):
      names = set(self.imports[module])
      if module == "typing":
        need_typing = False
        for (name, count) in self._typing_import_counts.items():
          if count:
            need_typing = True
          else:
            names.discard(name)
        if not need_typing:
          names.discard(None)
      if None in names:
        ret.append(f"import {module}")
        names.remove(None)

      if names:
        name_str = ", ".join(sorted(names))
        ret.append(f"from {module} import {name_str}")

    return ret

  def _IsBuiltin(self, module):
    return module == "builtins"

  def _FormatTypeParams(self, type_params):
    formatted_type_params = []
    for t in type_params:
      args = [f"'{t.name}'"]
      args += [self.Print(c) for c in t.constraints]
      if t.bound:
        args.append(f"bound={self.Print(t.bound)}")
      formatted_type_params.append(f"{t.name} = TypeVar({', '.join(args)})")
    return sorted(formatted_type_params)

  def _NameCollision(self, name):

    def name_in(members):
      return name in members or (
          self._unit and f"{self._unit.name}.{name}" in members)

    return name_in(self._class_members) or name_in(self._local_names)

  def _FromTyping(self, name):
    self._typing_import_counts[name] += 1
    if self._NameCollision(name):
      self._RequireImport("typing")
      return f"typing.{name}"
    else:
      self._RequireImport("typing", name)
      return name

  def _ImportTypingExtension(self, name):
    # `name` is a typing construct that is not supported in all Python versions.
    if self._local_names.get(name) == "alias":
      # A typing_extensions import is parsed as Alias(X, typing_extensions.X).
      # If we see an alias to `name`, assume it's been explicitly imported from
      # typing_extensions due to the current Python version not supporting it.
      return name
    else:
      return self._FromTyping(name)

  def _StripUnitPrefix(self, name):
    if self._unit:
      return utils.strip_prefix(name, f"{self._unit.name}.")
    else:
      return name

  def EnterTypeDeclUnit(self, unit):
    self._unit = unit
    for definitions, label in [(unit.classes, "class"),
                               (unit.functions, "function"),
                               (unit.constants, "constant"),
                               (unit.type_params, "type_param"),
                               (unit.aliases, "alias")]:
      for defn in definitions:
        self._local_names[defn.name] = label
    for alias in unit.aliases:
      # Modules are represented as NamedTypes in partially resolved asts and
      # sometimes as LateTypes in asts modified for pickling.
      if isinstance(alias.type, pytd.Module):
        module_name = alias.type.module_name
      elif isinstance(alias.type, (pytd.NamedType, pytd.LateType)):
        module_name = alias.type.name
      else:
        continue
      name = self._StripUnitPrefix(alias.name)
      self._module_aliases[module_name] = name

  def LeaveTypeDeclUnit(self, _):
    self._unit = None
    self._local_names = {}

  def VisitTypeDeclUnit(self, node):
    """Convert the AST for an entire module back to a string."""
    if node.type_params:
      self._FromTyping("TypeVar")

    aliases = []
    imports = set(self._GenerateImportStrings())
    for alias in node.aliases:
      if alias.startswith(("from ", "import ")):
        imports.add(alias)
      else:
        aliases.append(alias)

    # Sort import lines lexicographically and ensure import statements come
    # before from-import statements.
    imports = sorted(imports, key=lambda s: (s.startswith("from "), s))

    # Remove deleted constants
    constants = [c for c in node.constants if "<deleted>" not in c]

    sections = [
        imports,
        aliases,
        constants,
        self._FormatTypeParams(self.old_node.type_params),
        node.classes,
        node.functions,
    ]

    # We put one blank line after every class,so we need to strip the blank line
    # after the last class.
    sections_as_string = ("\n".join(section_suite).rstrip()
                          for section_suite in sections
                          if section_suite)
    return "\n\n".join(sections_as_string)

  def EnterConstant(self, node):
    self.in_constant = True

  def LeaveConstant(self, node):
    self.in_constant = False

  def _DropTypingConstant(self, node):
    # Hack to account for a corner case in late annotation handling.
    # If we have a top-level constant of the exact form
    #   Foo: Type[typing.Foo]
    # we drop the constant and rewrite it to
    #   from typing import Foo
    if self.class_names or node.value:
      return False
    if node.type == f"Type[typing.{node.name}]":
      self._RequireImport("typing", node.name)
      self._typing_import_counts["Type"] -= 1
      del self._local_names[node.name]
      return True

  def VisitConstant(self, node):
    """Convert a class-level or module-level constant to a string."""
    if self.in_literal:
      # This should be either True, False or an enum. For the booleans, strip
      # off the module name. For enums, print the whole name.
      if "builtins." in node.name:
        _, _, name = node.name.partition(".")
        return name
      else:
        return node.name
    # Decrement Any, since the actual value is never printed.
    if node.value == "Any":
      self._typing_import_counts["Any"] -= 1
    if self._DropTypingConstant(node):
      return "<deleted>"
    # Whether the constant has a default value is important for fields in
    # generated classes like namedtuples.
    suffix = " = ..." if node.value else ""
    return f"{node.name}: {node.type}{suffix}"

  def EnterAlias(self, _):
    self.old_imports = self.imports.copy()

  def VisitAlias(self, node):
    """Convert an import or alias to a string."""
    if (isinstance(self.old_node.type,
                   (pytd.NamedType, pytd.ClassType, pytd.LateType)) and
        not self.in_constant and not self.in_signature):
      full_name = self.old_node.type.name
      suffix = ""
      module, _, name = full_name.rpartition(".")
      if module:
        alias_name = self._StripUnitPrefix(self.old_node.name)
        if name not in ("*", alias_name):
          suffix += f" as {alias_name}"
        self.imports = self.old_imports  # undo unnecessary imports change
        return f"from {module} import {name}{suffix}"
    elif isinstance(self.old_node.type, (pytd.Constant, pytd.Function)):
      return self.Print(self.old_node.type.Replace(name=node.name))
    elif isinstance(self.old_node.type, pytd.Module):
      return node.type
    return f"{node.name} = {node.type}"

  def EnterClass(self, node):
    """Entering a class - record class name for children's use."""
    n = node.name
    if node.template:
      n += f"[{', '.join(self.Print(t) for t in node.template)}]"
    for member in node.methods + node.constants:
      self._class_members.add(member.name)
    self.class_names.append(n)
    # Class decorators are resolved to their underlying functions, but all we
    # output is '@{decorator.name}', so we do not want to visit the Function()
    # node and collect types etc. (In particular, we would add a spurious import
    # of 'Any' when generating a decorator for an InterpreterClass.)
    return {"decorators"}

  def LeaveClass(self, unused_node):
    self._class_members.clear()
    self.class_names.pop()

  def VisitClass(self, node):
    """Visit a class, producing a multi-line, properly indented string."""
    bases = node.bases
    # If object is the only base, we don't need to list any bases.
    if bases == ("object",):
      bases = ()
    if node.metaclass is not None:
      bases += ("metaclass=" + node.metaclass,)
    bases_str = f"({', '.join(bases)})" if bases else ""
    header = [f"class {node.name}{bases_str}:"]
    if node.slots is not None:
      slots_str = ", ".join(f"\"{s}\"" for s in node.slots)
      slots = [self.INDENT + f"__slots__ = [{slots_str}]"]
    else:
      slots = []
    decorators = ["@" + self.VisitNamedType(d)
                  for d in self.old_node.decorators]
    # Our handling of class decorators is a bit hacky (see output.py); this
    # makes sure that typing classes read in directly from a pyi file and then
    # reemitted (e.g. in assertTypesMatchPytd) have their required module
    # imports handled correctly.
    for d in self.old_node.decorators:
      if d.type.name.startswith("typing."):
        self.VisitNamedType(d.type)
    if node.classes or node.methods or node.constants or slots:
      # We have multiple methods, and every method has multiple signatures
      # (i.e., the method string will have multiple lines). Combine this into
      # an array that contains all the lines, then indent the result.
      class_lines = sum((m.splitlines() for m in node.classes), [])
      classes = [self.INDENT + m for m in class_lines]
      constants = [self.INDENT + m for m in node.constants]
      method_lines = sum((m.splitlines() for m in node.methods), [])
      methods = [self.INDENT + m for m in method_lines]
    else:
      header[-1] += " ..."
      constants = []
      classes = []
      methods = []
    lines = decorators + header + slots + classes + constants + methods
    return "\n".join(lines) + "\n"

  def VisitFunction(self, node):
    """Visit function, producing multi-line string (one for each signature)."""
    function_name = node.name
    decorators = ""
    if node.is_final:
      decorators += "@" + self._FromTyping("final") + "\n"
    if (node.kind == pytd.MethodKind.STATICMETHOD and
        function_name != "__new__"):
      decorators += "@staticmethod\n"
    elif (node.kind == pytd.MethodKind.CLASSMETHOD and
          function_name != "__init_subclass__"):
      decorators += "@classmethod\n"
    elif node.kind == pytd.MethodKind.PROPERTY:
      decorators += "@property\n"
    if node.is_abstract:
      decorators += "@abstractmethod\n"
    if node.is_coroutine:
      decorators += "@coroutine\n"
    if len(node.signatures) > 1:
      decorators += "@" + self._FromTyping("overload") + "\n"
    signatures = "\n".join(decorators + "def " + function_name + sig
                           for sig in node.signatures)
    return signatures

  def _FormatContainerContents(self, node):
    """Print out the last type parameter of a container. Used for *args/**kw."""
    assert isinstance(node, pytd.Parameter)
    if isinstance(node.type, pytd.GenericType):
      container_name = node.type.name.rpartition(".")[2]
      assert container_name in ("tuple", "dict")
      self._typing_import_counts[container_name.capitalize()] -= 1
      # If the type is "Any", e.g. `**kwargs: Any`, decrement Any to avoid an
      # extraneous import of typing.Any. Any was visited before this function
      # transformed **kwargs, so it was incremented at least once already.
      if isinstance(node.type.parameters[-1], pytd.AnythingType):
        self._typing_import_counts["Any"] -= 1
      return self.Print(
          node.Replace(type=node.type.parameters[-1], optional=False))
    else:
      return self.Print(node.Replace(type=pytd.AnythingType(), optional=False))

  def EnterSignature(self, node):
    self.in_signature = True

  def LeaveSignature(self, node):
    self.in_signature = False

  def VisitSignature(self, node):
    """Visit a signature, producing a string."""
    if node.return_type == "nothing":
      return_type = "NoReturn"  # a prettier alias for nothing
      self._FromTyping(return_type)
    else:
      return_type = node.return_type
    ret = f" -> {return_type}"

    # Put parameters in the right order:
    # (arg1, arg2, *args, kwonly1, kwonly2, **kwargs)
    if self.old_node.starargs is not None:
      starargs = self._FormatContainerContents(self.old_node.starargs)
    else:
      # We don't have explicit *args, but we might need to print "*", for
      # kwonly params.
      starargs = ""
    params = []
    for i, p in enumerate(node.params):
      if self.old_node.params[i].kind == pytd.ParameterKind.KWONLY:
        assert all(p.kind == pytd.ParameterKind.KWONLY
                   for p in self.old_node.params[i:])
        params.append("*" + starargs)
        params.extend(node.params[i:])
        break
      params.append(p)
      if (self.old_node.params[i].kind == pytd.ParameterKind.POSONLY and
          (i == len(node.params)-1 or
           self.old_node.params[i+1].kind != pytd.ParameterKind.POSONLY)):
        params.append("/")
    else:
      if starargs:
        params.append(f"*{starargs}")
    if self.old_node.starstarargs is not None:
      starstarargs = self._FormatContainerContents(self.old_node.starstarargs)
      params.append(f"**{starstarargs}")

    body = []
    # Handle Mutable parameters
    # pylint: disable=no-member
    # (old_node is set in parse/node.py)
    mutable_params = [(p.name, p.mutated_type) for p in self.old_node.params
                      if p.mutated_type is not None]
    # pylint: enable=no-member
    for name, new_type in mutable_params:
      body.append(f"\n{self.INDENT}{name} = {self.Print(new_type)}")
    for exc in node.exceptions:
      body.append(f"\n{self.INDENT}raise {exc}()")
    if not body:
      body.append(" ...")

    if self.multiline_args:
      indent = "\n" + self.INDENT
      params = ",".join([indent + p for p in params])
      return f"({params}\n){ret}:{''.join(body)}"
    else:
      params = ", ".join(params)
      return f"({params}){ret}:{''.join(body)}"

  def EnterParameter(self, unused_node):
    assert not self.in_parameter
    self.in_parameter = True

  def LeaveParameter(self, unused_node):
    assert self.in_parameter
    self.in_parameter = False

  def VisitParameter(self, node):
    """Convert a function parameter to a string."""
    suffix = " = ..." if node.optional else ""
    if node.type == "Any":
      # Abbreviated form. "Any" is the default.
      self._typing_import_counts["Any"] -= 1
      return node.name + suffix
    # For parameterized class, for example: ClsName[T, V].
    # Its name is `ClsName` before `[`.
    elif node.name == "self" and self.class_names and (
        self.class_names[-1].split("[")[0] == node.type.split("[")[0]):
      if "[" in node.type:
        elided = node.type.split("[", 1)[-1]
        for k in self._typing_import_counts:
          if re.search(r"(^|\W)%s($|\W)" % k, elided):
            self._typing_import_counts[k] -= 1
      return node.name + suffix
    elif node.name == "cls" and self.class_names and (
        node.type == f"Type[{self.class_names[-1]}]"):
      self._typing_import_counts["Type"] -= 1
      return node.name + suffix
    elif node.type is None:
      logging.warning("node.type is None")
      return node.name
    else:
      return node.name + ": " + node.type + suffix

  def VisitTemplateItem(self, node):
    """Convert a template to a string."""
    return node.type_param

  def _UseExistingModuleAlias(self, name):
    prefix, suffix = name.rsplit(".", 1)
    while prefix:
      if prefix in self._module_aliases:
        return f"{self._module_aliases[prefix]}.{suffix}"
      prefix, _, remainder = prefix.rpartition(".")
      suffix = f"{remainder}.{suffix}"
    return None

  def _GuessModule(self, maybe_module):
    """Guess which part of the given name is the module prefix."""
    if "." not in maybe_module:
      return maybe_module, ""
    prefix, suffix = maybe_module.rsplit(".", 1)
    # Heuristic: modules are typically lowercase, classes uppercase.
    if suffix[0].islower():
      return maybe_module, ""
    else:
      module, rest = self._GuessModule(prefix)
      return module, f"{rest}.{suffix}" if rest else suffix

  def VisitNamedType(self, node):
    """Convert a type to a string."""
    prefix, _, suffix = node.name.rpartition(".")
    if self._IsBuiltin(prefix) and not self._NameCollision(suffix):
      node_name = suffix
    elif prefix == "typing":
      node_name = self._FromTyping(suffix)
    elif "." not in node.name:
      node_name = node.name
    else:
      if self._unit:
        try:
          pytd.LookupItemRecursive(self._unit, self._StripUnitPrefix(node.name))
        except KeyError:
          aliased_name = self._UseExistingModuleAlias(node.name)
          if aliased_name:
            node_name = aliased_name
          else:
            module, rest = self._GuessModule(prefix)
            module_alias = module
            while self._NameCollision(module_alias):
              module_alias = f"_{module_alias}"
            if module_alias == module:
              self._RequireImport(module)
              node_name = node.name
            else:
              self._RequireImport(f"{module} as {module_alias}")
              node_name = ".".join(filter(bool, (module_alias, rest, suffix)))
        else:
          node_name = node.name
      else:
        node_name = node.name
    if node_name == "NoneType":
      # PEP 484 allows this special abbreviation.
      return "None"
    else:
      return node_name

  def VisitLateType(self, node):
    return self.VisitNamedType(node)

  def VisitClassType(self, node):
    return self.VisitNamedType(node)

  def VisitStrictType(self, node):
    # 'StrictType' is defined, and internally used, by booleq.py. We allow it
    # here so that booleq.py can use pytd_utils.Print().
    return self.VisitNamedType(node)

  def VisitAnythingType(self, unused_node):
    """Convert an anything type to a string."""
    return self._FromTyping("Any")

  def VisitNothingType(self, unused_node):
    """Convert the nothing type to a string."""
    return "nothing"

  def VisitTypeParameter(self, node):
    return node.name

  def VisitModule(self, node):
    if self.in_constant or self.in_signature:
      return "module"
    elif not node.is_aliased:
      return f"import {node.module_name}"
    elif "." in node.module_name:
      # `import x.y as z` and `from x import y as z` are equivalent, but the
      # latter is a bit prettier.
      prefix, suffix = node.module_name.rsplit(".", 1)
      imp = f"from {prefix} import {suffix}"
      if node.name != suffix:
        imp += f" as {node.name}"
      return imp
    else:
      return f"import {node.module_name} as {node.name}"

  def MaybeCapitalize(self, name):
    """Capitalize a generic type, if necessary."""
    capitalized = pep484.PEP484_MaybeCapitalize(name)
    if capitalized:
      return self._FromTyping(capitalized)
    else:
      return name

  def VisitGenericType(self, node):
    """Convert a generic type to a string."""
    parameters = node.parameters
    if self._IsEmptyTuple(node):
      parameters = ("()",)
    elif self._NeedsTupleEllipsis(node):
      parameters += ("...",)
    elif self._NeedsCallableEllipsis(self.old_node):
      # Callable[Any, X] is rewritten to Callable[..., X].
      self._typing_import_counts["Any"] -= 1
      parameters = ("...",) + parameters[1:]
    return (self.MaybeCapitalize(node.base_type) +
            "[" + ", ".join(parameters) + "]")

  def VisitCallableType(self, node):
    typ = self.MaybeCapitalize(node.base_type)
    args = ", ".join(node.args)
    return f"{typ}[[{args}], {node.ret}]"

  def VisitTupleType(self, node):
    return self.VisitGenericType(node)

  def VisitUnionType(self, node):
    """Convert a union type ("x or y") to a string."""
    type_list = self._FormSetTypeList(node)
    return self._BuildUnion(type_list)

  def VisitIntersectionType(self, node):
    """Convert a intersection type ("x and y") to a string."""
    type_list = self._FormSetTypeList(node)
    return self._BuildIntersection(type_list)

  def _FormSetTypeList(self, node):
    """Form list of types within a set type."""
    type_list = dict.fromkeys(node.type_list)
    if self.in_parameter:
      # Parameter's set types are merged after as a follow up to the
      # ExpandCompatibleBuiltins visitor.
      for compat, name in pep484.COMPAT_ITEMS:
        # name can replace compat.
        if compat in type_list and name in type_list:
          del type_list[compat]
    return type_list

  def _BuildUnion(self, type_list):
    """Builds a union of the types in type_list.

    Args:
      type_list: A list of strings representing types.

    Returns:
      A string representing the union of the types in type_list. Simplifies
      Union[X] to X and Union[X, None] to Optional[X].
    """
    # Collect all literals, so we can print them using the Literal[x1, ..., xn]
    # syntactic sugar.
    literals = []
    new_type_list = []
    for t in type_list:
      match = re.fullmatch(r"Literal\[(?P<content>.*)\]", t)
      if match:
        literals.append(match.group("content"))
      else:
        new_type_list.append(t)
    if literals:
      new_type_list.append(f"Literal[{', '.join(literals)}]")
    if len(new_type_list) == 1:
      return new_type_list[0]
    elif "None" in new_type_list:
      return (self._FromTyping("Optional") + "[" +
              self._BuildUnion(t for t in new_type_list if t != "None") + "]")
    else:
      return self._FromTyping("Union") + "[" + ", ".join(new_type_list) + "]"

  def _BuildIntersection(self, type_list):
    """Builds a intersection of the types in type_list.

    Args:
      type_list: A list of strings representing types.

    Returns:
      A string representing the intersection of the types in type_list.
      Simplifies Intersection[X] to X and Intersection[X, None] to Optional[X].
    """
    type_list = tuple(type_list)
    if len(type_list) == 1:
      return type_list[0]
    else:
      return " and ".join(type_list)

  def EnterLiteral(self, _):
    assert not self.in_literal
    self.in_literal = True

  def LeaveLiteral(self, _):
    assert self.in_literal
    self.in_literal = False

  def VisitLiteral(self, node):
    base = self._ImportTypingExtension("Literal")
    return f"{base}[{node.value}]"

  def VisitAnnotated(self, node):
    base = self._ImportTypingExtension("Annotated")
    annotations = ", ".join(node.annotations)
    return f"{base}[{node.base_type}, {annotations}]"



from typing import AnyStr

class error(Exception):
    def __init__(self, message: str, pattern: AnyStr | None = ..., pos: int | None = ...) -> None: ...

A: int
ASCII: int
B: int
BESTMATCH: int
D: int
DEBUG: int
E: int
ENHANCEMATCH: int
F: int
FULLCASE: int
I: int
IGNORECASE: int
L: int
LOCALE: int
M: int
MULTILINE: int
P: int
POSIX: int
R: int
REVERSE: int
T: int
TEMPLATE: int
S: int
DOTALL: int
U: int
UNICODE: int
V0: int
VERSION0: int
V1: int
VERSION1: int
W: int
WORD: int
X: int
VERBOSE: int

DEFAULT_VERSION: int



from typing import Any, Generic, TypeVar

from .. import util
from . import operators
from .base import SchemaEventTarget
from .visitors import Traversible, TraversibleType

_T = TypeVar("_T")

BOOLEANTYPE: Any
INTEGERTYPE: Any
NULLTYPE: Any
STRINGTYPE: Any
MATCHTYPE: Any
INDEXABLE: Any
TABLEVALUE: Any

class TypeEngine(Traversible):
    class Comparator(operators.ColumnOperators[_T], Generic[_T]):
        default_comparator: Any
        def __clause_element__(self): ...
        expr: Any
        type: Any
        def __init__(self, expr) -> None: ...
        def operate(self, op, *other, **kwargs): ...
        def reverse_operate(self, op, other, **kwargs): ...
        def __reduce__(self): ...
    hashable: bool
    comparator_factory: Any
    sort_key_function: Any
    should_evaluate_none: bool
    def evaluates_none(self): ...
    def copy(self, **kw): ...
    def compare_against_backend(self, dialect, conn_type) -> None: ...
    def copy_value(self, value): ...
    def literal_processor(self, dialect) -> None: ...
    def bind_processor(self, dialect) -> None: ...
    def result_processor(self, dialect, coltype) -> None: ...
    def column_expression(self, colexpr) -> None: ...
    def bind_expression(self, bindvalue) -> None: ...
    def compare_values(self, x, y): ...
    def get_dbapi_type(self, dbapi) -> None: ...
    @property
    def python_type(self) -> None: ...
    def with_variant(self, type_, dialect_name): ...
    def as_generic(self, allow_nulltype: bool = ...): ...
    def dialect_impl(self, dialect): ...
    def adapt(self, cls, **kw): ...
    def coerce_compared_value(self, op, value): ...
    def compile(self, dialect: Any | None = ...): ...

class VisitableCheckKWArg(util.EnsureKWArgType, TraversibleType): ...

class ExternalType:
    cache_ok: Any

class UserDefinedType:
    __visit_name__: str
    ensure_kwarg: str
    def coerce_compared_value(self, op, value): ...

class Emulated:
    def adapt_to_emulated(self, impltype, **kw): ...
    def adapt(self, impltype, **kw): ...

class NativeForEmulated:
    @classmethod
    def adapt_native_to_emulated(cls, impl, **kw): ...
    @classmethod
    def adapt_emulated_to_native(cls, impl, **kw): ...

class TypeDecorator(ExternalType, SchemaEventTarget, TypeEngine):
    __visit_name__: str
    impl: Any
    def __init__(self, *args, **kwargs) -> None: ...
    coerce_to_is_types: Any

    class Comparator(TypeEngine.Comparator):
        def operate(self, op, *other, **kwargs): ...
        def reverse_operate(self, op, other, **kwargs): ...

    @property
    def comparator_factory(self): ...
    def type_engine(self, dialect): ...
    def load_dialect_impl(self, dialect): ...
    def __getattr__(self, key): ...
    def process_literal_param(self, value, dialect) -> None: ...
    def process_bind_param(self, value, dialect) -> None: ...
    def process_result_value(self, value, dialect) -> None: ...
    def literal_processor(self, dialect): ...
    def bind_processor(self, dialect): ...
    def result_processor(self, dialect, coltype): ...
    def bind_expression(self, bindparam): ...
    def column_expression(self, column): ...
    def coerce_compared_value(self, op, value): ...
    def copy(self, **kw): ...
    def get_dbapi_type(self, dbapi): ...
    def compare_values(self, x, y): ...
    @property
    def sort_key_function(self): ...

class Variant(TypeDecorator):
    cache_ok: bool
    impl: Any
    mapping: Any
    def __init__(self, base, mapping) -> None: ...
    def coerce_compared_value(self, operator, value): ...
    def load_dialect_impl(self, dialect): ...
    def with_variant(self, type_, dialect_name): ...
    @property
    def comparator_factory(self): ...

def to_instance(typeobj, *arg, **kw): ...
def adapt_type(typeobj, colspecs): ...



import sys
from typing import Any, ClassVar, Mapping

if sys.version_info >= (3, 9):
    __all__ = ["Dialog"]

class Dialog:
    command: ClassVar[str | None]
    master: Any | None
    options: Mapping[str, Any]
    def __init__(self, master: Any | None = ..., **options) -> None: ...
    def show(self, **options) -> Any: ...



from typing import ClassVar

import passlib.utils.handlers as uh

class cisco_pix(uh.HasUserContext, uh.StaticHandler):
    name: ClassVar[str]
    truncate_size: ClassVar[int]
    truncate_error: ClassVar[bool]
    truncate_verify_reject: ClassVar[bool]
    checksum_size: ClassVar[int]
    checksum_chars: ClassVar[str]

class cisco_asa(cisco_pix): ...

class cisco_type7(uh.GenericHandler):
    name: ClassVar[str]
    checksum_chars: ClassVar[str]
    min_salt_value: ClassVar[int]
    max_salt_value: ClassVar[int]
    @classmethod
    def using(cls, salt: int | None = ..., **kwds): ...  # type: ignore[override]
    @classmethod
    def from_string(cls, hash): ...
    salt: int
    def __init__(self, salt: int | None = ..., **kwds) -> None: ...
    @classmethod
    def decode(cls, hash, encoding: str = ...): ...



class Extension:
    def __init__(
        self,
        name: str,
        sources: list[str],
        include_dirs: list[str] = ...,
        define_macros: list[tuple[str, str | None]] = ...,
        undef_macros: list[str] = ...,
        library_dirs: list[str] = ...,
        libraries: list[str] = ...,
        runtime_library_dirs: list[str] = ...,
        extra_objects: list[str] = ...,
        extra_compile_args: list[str] = ...,
        extra_link_args: list[str] = ...,
        export_symbols: list[str] = ...,
        swig_opts: str | None = ...,  # undocumented
        depends: list[str] = ...,
        language: str = ...,
    ) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.well_known_types
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Duration(google.protobuf.message.Message, google.protobuf.internal.well_known_types.Duration):
    """A Duration represents a signed, fixed-length span of time represented
    as a count of seconds and fractions of seconds at nanosecond
    resolution. It is independent of any calendar and concepts like "day"
    or "month". It is related to Timestamp in that the difference between
    two Timestamp values is a Duration and it can be added or subtracted
    from a Timestamp. Range is approximately +-10,000 years.

    # Examples

    Example 1: Compute Duration from two Timestamps in pseudo code.

        Timestamp start = ...;
        Timestamp end = ...;
        Duration duration = ...;

        duration.seconds = end.seconds - start.seconds;
        duration.nanos = end.nanos - start.nanos;

        if (duration.seconds < 0 && duration.nanos > 0) {
          duration.seconds += 1;
          duration.nanos -= 1000000000;
        } else if (duration.seconds > 0 && duration.nanos < 0) {
          duration.seconds -= 1;
          duration.nanos += 1000000000;
        }

    Example 2: Compute Timestamp from Timestamp + Duration in pseudo code.

        Timestamp start = ...;
        Duration duration = ...;
        Timestamp end = ...;

        end.seconds = start.seconds + duration.seconds;
        end.nanos = start.nanos + duration.nanos;

        if (end.nanos < 0) {
          end.seconds -= 1;
          end.nanos += 1000000000;
        } else if (end.nanos >= 1000000000) {
          end.seconds += 1;
          end.nanos -= 1000000000;
        }

    Example 3: Compute Duration from datetime.timedelta in Python.

        td = datetime.timedelta(days=3, minutes=10)
        duration = Duration()
        duration.FromTimedelta(td)

    # JSON Mapping

    In JSON format, the Duration type is encoded as a string rather than an
    object, where the string ends in the suffix "s" (indicating seconds) and
    is preceded by the number of seconds, with nanoseconds expressed as
    fractional seconds. For example, 3 seconds with 0 nanoseconds should be
    encoded in JSON format as "3s", while 3 seconds and 1 nanosecond should
    be expressed in JSON format as "3.000000001s", and 3 seconds and 1
    microsecond should be expressed in JSON format as "3.000001s".
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SECONDS_FIELD_NUMBER: builtins.int
    NANOS_FIELD_NUMBER: builtins.int
    seconds: builtins.int
    """Signed seconds of the span of time. Must be from -315,576,000,000
    to +315,576,000,000 inclusive. Note: these bounds are computed from:
    60 sec/min * 60 min/hr * 24 hr/day * 365.25 days/year * 10000 years
    """

    nanos: builtins.int
    """Signed fractions of a second at nanosecond resolution of the span
    of time. Durations less than one second are represented with a 0
    `seconds` field and a positive or negative `nanos` field. For durations
    of one second or more, a non-zero value for the `nanos` field must be
    of the same sign as the `seconds` field. Must be from -999,999,999
    to +999,999,999 inclusive.
    """

    def __init__(self,
        *,
        seconds: typing.Optional[builtins.int] = ...,
        nanos: typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["nanos",b"nanos","seconds",b"seconds"]) -> None: ...
global___Duration = Duration



import threading
from typing import Iterable

class StreamWatcher(threading.local):
    def submit(self, stream) -> Iterable[str]: ...

class Responder(StreamWatcher):
    pattern: str
    response: str
    index: int
    def __init__(self, pattern: str, response: str) -> None: ...
    def pattern_matches(self, stream: str, pattern: str, index_attr: str) -> Iterable[str]: ...
    def submit(self, stream: str) -> Iterable[str]: ...

class FailingResponder(Responder):
    sentinel: str
    failure_index: int
    tried: bool
    def __init__(self, pattern: str, response: str, sentinel: str) -> None: ...
    def submit(self, stream: str) -> Iterable[str]: ...



from typing import Any

from .geometry import Point2D, PositiveSize2D, Transform2D

class XDRPoint2D(Point2D):
    namespace: Any
    x: Any
    y: Any

class XDRPositiveSize2D(PositiveSize2D):
    namespace: Any
    cx: Any
    cy: Any

class XDRTransform2D(Transform2D):
    namespace: Any
    rot: Any
    flipH: Any
    flipV: Any
    off: Any
    ext: Any
    chOff: Any
    chExt: Any



import sys

from .case import (
    FunctionTestCase as FunctionTestCase,
    SkipTest as SkipTest,
    TestCase as TestCase,
    expectedFailure as expectedFailure,
    skip as skip,
    skipIf as skipIf,
    skipUnless as skipUnless,
)
from .loader import (
    TestLoader as TestLoader,
    defaultTestLoader as defaultTestLoader,
    findTestCases as findTestCases,
    getTestCaseNames as getTestCaseNames,
    makeSuite as makeSuite,
)
from .main import TestProgram as TestProgram, main as main
from .result import TestResult as TestResult
from .runner import TextTestResult as TextTestResult, TextTestRunner as TextTestRunner
from .signals import (
    installHandler as installHandler,
    registerResult as registerResult,
    removeHandler as removeHandler,
    removeResult as removeResult,
)
from .suite import BaseTestSuite as BaseTestSuite, TestSuite as TestSuite

if sys.version_info >= (3, 8):
    from unittest.async_case import *

    from .case import addModuleCleanup as addModuleCleanup

    __all__ = [
        "TestResult",
        "TestCase",
        "IsolatedAsyncioTestCase",
        "TestSuite",
        "TextTestRunner",
        "TestLoader",
        "FunctionTestCase",
        "main",
        "defaultTestLoader",
        "SkipTest",
        "skip",
        "skipIf",
        "skipUnless",
        "expectedFailure",
        "TextTestResult",
        "installHandler",
        "registerResult",
        "removeResult",
        "removeHandler",
        "addModuleCleanup",
        "getTestCaseNames",
        "makeSuite",
        "findTestCases",
    ]

else:
    __all__ = [
        "TestResult",
        "TestCase",
        "TestSuite",
        "TextTestRunner",
        "TestLoader",
        "FunctionTestCase",
        "main",
        "defaultTestLoader",
        "SkipTest",
        "skip",
        "skipIf",
        "skipUnless",
        "expectedFailure",
        "TextTestResult",
        "installHandler",
        "registerResult",
        "removeResult",
        "removeHandler",
        "getTestCaseNames",
        "makeSuite",
        "findTestCases",
    ]

def load_tests(loader: TestLoader, tests: TestSuite, pattern: str | None) -> TestSuite: ...



from collections import defaultdict
from typing import Any

class BoundDictionary(defaultdict[Any, Any]):
    reference: Any
    def __init__(self, reference: Any | None = ..., *args, **kw) -> None: ...
    def __getitem__(self, key): ...



from typing import Any

from braintree.error_result import ErrorResult as ErrorResult
from braintree.exceptions.not_found_error import NotFoundError as NotFoundError
from braintree.exceptions.request_timeout_error import RequestTimeoutError as RequestTimeoutError
from braintree.resource import Resource as Resource
from braintree.resource_collection import ResourceCollection as ResourceCollection
from braintree.successful_result import SuccessfulResult as SuccessfulResult
from braintree.transaction import Transaction as Transaction

class TransactionGateway:
    gateway: Any
    config: Any
    def __init__(self, gateway) -> None: ...
    def adjust_authorization(self, transaction_id, amount): ...
    def clone_transaction(self, transaction_id, params): ...
    def cancel_release(self, transaction_id): ...
    def create(self, params): ...
    def credit(self, params): ...
    def find(self, transaction_id): ...
    def hold_in_escrow(self, transaction_id): ...
    def refund(self, transaction_id, amount_or_options: Any | None = ...): ...
    def sale(self, params): ...
    def search(self, *query): ...
    def release_from_escrow(self, transaction_id): ...
    def submit_for_settlement(self, transaction_id, amount: Any | None = ..., params: Any | None = ...): ...
    def update_details(self, transaction_id, params: Any | None = ...): ...
    def submit_for_partial_settlement(self, transaction_id, amount, params: Any | None = ...): ...
    def void(self, transaction_id): ...



from typing import Any

class ReverseDnsSetting:
    OFF: Any
    REQUIRE_RESOLVE_ALL_ADDRESSES: Any
    REQUIRE_RESOLVE_IP_ADDRESSES_ONLY: Any
    OPTIONAL_RESOLVE_ALL_ADDRESSES: Any
    OPTIONAL_RESOLVE_IP_ADDRESSES_ONLY: Any
    SUPPORTED_VALUES: Any

def get_hostname_by_addr(addr, success_required: bool = ...): ...
def is_ip_addr(addr): ...



from codeop import CommandCompiler
from types import CodeType
from typing import Any, Callable, Mapping

__all__ = ["InteractiveInterpreter", "InteractiveConsole", "interact", "compile_command"]

class InteractiveInterpreter:
    locals: Mapping[str, Any]  # undocumented
    compile: CommandCompiler  # undocumented
    def __init__(self, locals: Mapping[str, Any] | None = ...) -> None: ...
    def runsource(self, source: str, filename: str = ..., symbol: str = ...) -> bool: ...
    def runcode(self, code: CodeType) -> None: ...
    def showsyntaxerror(self, filename: str | None = ...) -> None: ...
    def showtraceback(self) -> None: ...
    def write(self, data: str) -> None: ...

class InteractiveConsole(InteractiveInterpreter):
    buffer: list[str]  # undocumented
    filename: str  # undocumented
    def __init__(self, locals: Mapping[str, Any] | None = ..., filename: str = ...) -> None: ...
    def interact(self, banner: str | None = ..., exitmsg: str | None = ...) -> None: ...
    def push(self, line: str) -> bool: ...
    def resetbuffer(self) -> None: ...
    def raw_input(self, prompt: str = ...) -> str: ...

def interact(
    banner: str | None = ...,
    readfunc: Callable[[str], str] | None = ...,
    local: Mapping[str, Any] | None = ...,
    exitmsg: str | None = ...,
) -> None: ...
def compile_command(source: str, filename: str = ..., symbol: str = ...) -> CodeType | None: ...



"""Test state.py."""

from pytype import compare
from pytype import state
from pytype.typegraph import cfg

import unittest


def source_summary(binding, **varnames):
  """A simple deterministic listing of source variables."""
  clauses = []
  name_map = {b.variable: name for name, b in varnames.items()}
  for origin in binding.origins:
    for sources in origin.source_sets:
      bindings = [f"{name_map[b.variable]}={b.data}" for b in sources]
      clauses.append(" ".join(sorted(bindings)))
  return " | ".join(sorted(clauses))


class FakeValue:

  def __init__(self, name, true_compat, false_compat):
    self._name = name
    self.compatible = {
        True: true_compat,
        False: false_compat}

  def __str__(self):
    return self._name


ONLY_TRUE = FakeValue("T", True, False)
ONLY_FALSE = FakeValue("F", False, True)
AMBIGUOUS = FakeValue("?", True, True)


def fake_compatible_with(value, logical_value):
  return value.compatible[logical_value]


class ConditionTestBase(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self._program = cfg.Program()
    self._node = self._program.NewCFGNode("test")
    self._old_compatible_with = compare.compatible_with
    compare.compatible_with = fake_compatible_with

  def tearDown(self):
    super().tearDown()
    compare.compatible_with = self._old_compatible_with

  def new_binding(self, value=AMBIGUOUS):
    var = self._program.NewVariable()
    return var.AddBinding(value)

  def check_binding(self, expected, binding, **varnames):
    self.assertEqual(len(binding.origins), 1)
    self.assertEqual(self._node, binding.origins[0].where)
    self.assertEqual(expected, source_summary(binding, **varnames))


class ConditionTest(ConditionTestBase):

  def test_no_parent(self):
    x = self.new_binding()
    y = self.new_binding()
    z = self.new_binding()
    c = state.Condition(self._node, [[x, y], [z]])
    self.check_binding("x=? y=? | z=?", c.binding, x=x, y=y, z=z)

  def test_parent_combination(self):
    p = self.new_binding()
    x = self.new_binding()
    y = self.new_binding()
    z = self.new_binding()
    c = state.Condition(self._node, [[x, y], [z]])
    self.check_binding("x=? y=? | z=?", c.binding,
                       p=p, x=x, y=y, z=z)


class SplitConditionTest(ConditionTestBase):

  def test(self):
    # Test that we split both sides and that everything gets passed through
    # correctly.  Don't worry about special cases within _restrict_condition
    # since those are tested separately.
    self.new_binding()
    var = self._program.NewVariable()
    var.AddBinding(ONLY_TRUE)
    var.AddBinding(ONLY_FALSE)
    var.AddBinding(AMBIGUOUS)
    true_cond, false_cond = state.split_conditions(self._node, var)
    self.check_binding("v=? | v=T", true_cond.binding,
                       v=var.bindings[0])
    self.check_binding("v=? | v=F",
                       false_cond.binding,
                       v=var.bindings[0])


class RestrictConditionTest(ConditionTestBase):

  def setUp(self):
    super().setUp()
    p = self.new_binding()
    self._parent = state.Condition(self._node, [[p]])

  def test_no_bindings(self):
    c = state._restrict_condition(self._node, [], False)
    self.assertIs(state.UNSATISFIABLE, c)
    c = state._restrict_condition(self._node, [], True)
    self.assertIs(state.UNSATISFIABLE, c)

  def test_none_restricted(self):
    x = self.new_binding()
    y = self.new_binding()
    state._restrict_condition(self._node, [x, y], False)
    state._restrict_condition(self._node, [x, y], True)

  def test_all_restricted(self):
    x = self.new_binding(ONLY_FALSE)
    y = self.new_binding(ONLY_FALSE)
    c = state._restrict_condition(self._node, [x, y], True)
    self.assertIs(state.UNSATISFIABLE, c)

  def test_some_restricted_no_parent(self):
    x = self.new_binding()  # Can be true or false.
    y = self.new_binding(ONLY_FALSE)
    z = self.new_binding()  # Can be true or false.
    c = state._restrict_condition(self._node, [x, y, z], True)
    self.check_binding("x=? | z=?", c.binding, x=x, y=y, z=z)

  def test_some_restricted_with_parent(self):
    x = self.new_binding()  # Can be true or false.
    y = self.new_binding(ONLY_FALSE)
    z = self.new_binding()  # Can be true or false.
    c = state._restrict_condition(self._node, [x, y, z], True)
    self.check_binding("x=? | z=?", c.binding,
                       x=x, y=y, z=z)

  def test_restricted_to_dnf(self):
    # DNF for a | (b & c)
    a = self.new_binding()
    b = self.new_binding()
    c = self.new_binding()
    dnf = [[a],
           [b, c]]
    x = self.new_binding()  # Compatible with everything
    y = self.new_binding(FakeValue("DNF", dnf, False))  # Reduce to dnf
    cond = state._restrict_condition(self._node, [x, y], True)
    self.check_binding("a=? | b=? c=? | x=?", cond.binding,
                       a=a, b=b, c=c, x=x, y=y)


if __name__ == "__main__":
  unittest.main()



from typing import Any

class InputDevice:
    name: Any
    actions: Any
    def __init__(self, name: Any | None = ...) -> None: ...
    def add_action(self, action) -> None: ...
    def clear_actions(self) -> None: ...
    def create_pause(self, duraton: int = ...) -> None: ...



"""Tests for pytype_source_utils.py."""

import os

from pytype import pytype_source_utils

import unittest


class PytypeSourceUtilsTest(unittest.TestCase):
  """Test pytype source utilities."""

  def setUp(self):
    super().setUp()
    self.root = os.path.dirname(__file__)

  def test_pytype_source_dir(self):
    self.assertEqual(self.root, pytype_source_utils.pytype_source_dir())

  def test_get_full_path(self):
    self.assertEqual(
        os.path.join(self.root, "foo/bar"),
        pytype_source_utils.get_full_path("foo/bar"))
    self.assertEqual(
        "/foo/bar",
        pytype_source_utils.get_full_path("/foo/bar"))

  def test_list_pytype_files(self):
    l = list(pytype_source_utils.list_pytype_files("stubs/stdlib"))
    self.assertIn("_ctypes.pytd", l)
    self.assertIn("collections/__init__.pytd", l)


if __name__ == "__main__":
  unittest.main()



import subprocess
from collections import deque
from typing import IO, Any, Callable, Sequence

from . import events, futures, protocols, transports

_File = int | IO[Any] | None

class BaseSubprocessTransport(transports.SubprocessTransport):

    _closed: bool  # undocumented
    _protocol: protocols.SubprocessProtocol  # undocumented
    _loop: events.AbstractEventLoop  # undocumented
    _proc: subprocess.Popen[Any] | None  # undocumented
    _pid: int | None  # undocumented
    _returncode: int | None  # undocumented
    _exit_waiters: list[futures.Future[Any]]  # undocumented
    _pending_calls: deque[tuple[Callable[..., Any], tuple[Any, ...]]]  # undocumented
    _pipes: dict[int, _File]  # undocumented
    _finished: bool  # undocumented
    def __init__(
        self,
        loop: events.AbstractEventLoop,
        protocol: protocols.SubprocessProtocol,
        args: str | bytes | Sequence[str | bytes],
        shell: bool,
        stdin: _File,
        stdout: _File,
        stderr: _File,
        bufsize: int,
        waiter: futures.Future[Any] | None = ...,
        extra: Any | None = ...,
        **kwargs: Any,
    ) -> None: ...
    def _start(
        self,
        args: str | bytes | Sequence[str | bytes],
        shell: bool,
        stdin: _File,
        stdout: _File,
        stderr: _File,
        bufsize: int,
        **kwargs: Any,
    ) -> None: ...  # undocumented
    def set_protocol(self, protocol: protocols.BaseProtocol) -> None: ...
    def get_protocol(self) -> protocols.BaseProtocol: ...
    def is_closing(self) -> bool: ...
    def close(self) -> None: ...
    def get_pid(self) -> int | None: ...  # type: ignore[override]
    def get_returncode(self) -> int | None: ...
    def get_pipe_transport(self, fd: int) -> _File: ...  # type: ignore[override]
    def _check_proc(self) -> None: ...  # undocumented
    def send_signal(self, signal: int) -> None: ...  # type: ignore[override]
    def terminate(self) -> None: ...
    def kill(self) -> None: ...
    async def _connect_pipes(self, waiter: futures.Future[Any] | None) -> None: ...  # undocumented
    def _call(self, cb: Callable[..., Any], *data: Any) -> None: ...  # undocumented
    def _pipe_connection_lost(self, fd: int, exc: BaseException | None) -> None: ...  # undocumented
    def _pipe_data_received(self, fd: int, data: bytes) -> None: ...  # undocumented
    def _process_exited(self, returncode: int) -> None: ...  # undocumented
    async def _wait(self) -> int: ...  # undocumented
    def _try_finish(self) -> None: ...  # undocumented
    def _call_connection_lost(self, exc: BaseException | None) -> None: ...  # undocumented

class WriteSubprocessPipeProto(protocols.BaseProtocol):  # undocumented
    def __init__(self, proc: BaseSubprocessTransport, fd: int) -> None: ...
    def connection_made(self, transport: transports.BaseTransport) -> None: ...
    def connection_lost(self, exc: BaseException | None) -> None: ...
    def pause_writing(self) -> None: ...
    def resume_writing(self) -> None: ...

class ReadSubprocessPipeProto(WriteSubprocessPipeProto, protocols.Protocol):  # undocumented
    def data_received(self, data: bytes) -> None: ...



from typing import Any

class StaticModule:
    def __init__(self, name) -> None: ...
    def __getattr__(self, attr): ...

def patch_path(path) -> None: ...
def read_configuration(filepath, find_others: bool = ..., ignore_option_errors: bool = ...): ...
def configuration_to_dict(handlers): ...
def parse_configuration(distribution, command_options, ignore_option_errors: bool = ...): ...

class ConfigHandler:
    section_prefix: Any
    aliases: Any
    ignore_option_errors: Any
    target_obj: Any
    sections: Any
    set_options: Any
    def __init__(self, target_obj, options, ignore_option_errors: bool = ...) -> None: ...
    @property
    def parsers(self) -> None: ...
    def __setitem__(self, option_name, value) -> None: ...
    def parse_section(self, section_options) -> None: ...
    def parse(self) -> None: ...

class ConfigMetadataHandler(ConfigHandler):
    section_prefix: str
    aliases: Any
    strict_mode: bool
    package_dir: Any
    def __init__(self, target_obj, options, ignore_option_errors: bool = ..., package_dir: Any | None = ...) -> None: ...
    @property
    def parsers(self): ...

class ConfigOptionsHandler(ConfigHandler):
    section_prefix: str
    @property
    def parsers(self): ...
    def parse_section_packages__find(self, section_options): ...
    def parse_section_entry_points(self, section_options) -> None: ...
    def parse_section_package_data(self, section_options) -> None: ...
    def parse_section_exclude_package_data(self, section_options) -> None: ...
    def parse_section_extras_require(self, section_options) -> None: ...
    def parse_section_data_files(self, section_options) -> None: ...



import collections.abc as _abc
from typing import Any

def normalize_locale(name): ...
def resolve_locale_filename(name): ...
def exists(name): ...
def locale_identifiers(): ...
def load(name, merge_inherited: bool = ...): ...
def merge(dict1, dict2) -> None: ...

class Alias:
    keys: Any
    def __init__(self, keys) -> None: ...
    def resolve(self, data): ...

class LocaleDataDict(_abc.MutableMapping):
    base: Any
    def __init__(self, data, base: Any | None = ...) -> None: ...
    def __len__(self): ...
    def __iter__(self): ...
    def __getitem__(self, key): ...
    def __setitem__(self, key, value) -> None: ...
    def __delitem__(self, key) -> None: ...
    def copy(self): ...



from collections.abc import Sequence
from typing import Pattern

def remove(path: str) -> None: ...
def findall_regex(items: Sequence[str], regex: Pattern[str]) -> list[int]: ...
def split_by_regex(items: Sequence[str], regex: Pattern[str]) -> list[Sequence[str]]: ...
def which(program: str) -> str | None: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class SheetBackgroundPicture(Serialisable):
    tagname: str
    id: Any
    def __init__(self, id) -> None: ...

class DrawingHF(Serialisable):
    id: Any
    lho: Any
    leftHeaderOddPages: Any
    lhe: Any
    leftHeaderEvenPages: Any
    lhf: Any
    leftHeaderFirstPage: Any
    cho: Any
    centerHeaderOddPages: Any
    che: Any
    centerHeaderEvenPages: Any
    chf: Any
    centerHeaderFirstPage: Any
    rho: Any
    rightHeaderOddPages: Any
    rhe: Any
    rightHeaderEvenPages: Any
    rhf: Any
    rightHeaderFirstPage: Any
    lfo: Any
    leftFooterOddPages: Any
    lfe: Any
    leftFooterEvenPages: Any
    lff: Any
    leftFooterFirstPage: Any
    cfo: Any
    centerFooterOddPages: Any
    cfe: Any
    centerFooterEvenPages: Any
    cff: Any
    centerFooterFirstPage: Any
    rfo: Any
    rightFooterOddPages: Any
    rfe: Any
    rightFooterEvenPages: Any
    rff: Any
    rightFooterFirstPage: Any
    def __init__(
        self,
        id: Any | None = ...,
        lho: Any | None = ...,
        lhe: Any | None = ...,
        lhf: Any | None = ...,
        cho: Any | None = ...,
        che: Any | None = ...,
        chf: Any | None = ...,
        rho: Any | None = ...,
        rhe: Any | None = ...,
        rhf: Any | None = ...,
        lfo: Any | None = ...,
        lfe: Any | None = ...,
        lff: Any | None = ...,
        cfo: Any | None = ...,
        cfe: Any | None = ...,
        cff: Any | None = ...,
        rfo: Any | None = ...,
        rfe: Any | None = ...,
        rff: Any | None = ...,
    ) -> None: ...



from typing import Any

from google.cloud.ndb import model

class _ClassKeyProperty(model.StringProperty):
    def __init__(self, name=..., indexed: bool = ...) -> None: ...

class PolyModel(model.Model):
    class_: Any



from typing import AnyStr, Callable, Generic, Iterable, Sequence, Text

DEFAULT_IGNORES: list[str]

def cmp(f1: bytes | Text, f2: bytes | Text, shallow: int | bool = ...) -> bool: ...
def cmpfiles(
    a: AnyStr, b: AnyStr, common: Iterable[AnyStr], shallow: int | bool = ...
) -> tuple[list[AnyStr], list[AnyStr], list[AnyStr]]: ...

class dircmp(Generic[AnyStr]):
    def __init__(
        self, a: AnyStr, b: AnyStr, ignore: Sequence[AnyStr] | None = ..., hide: Sequence[AnyStr] | None = ...
    ) -> None: ...
    left: AnyStr
    right: AnyStr
    hide: Sequence[AnyStr]
    ignore: Sequence[AnyStr]
    # These properties are created at runtime by __getattr__
    subdirs: dict[AnyStr, dircmp[AnyStr]]
    same_files: list[AnyStr]
    diff_files: list[AnyStr]
    funny_files: list[AnyStr]
    common_dirs: list[AnyStr]
    common_files: list[AnyStr]
    common_funny: list[AnyStr]
    common: list[AnyStr]
    left_only: list[AnyStr]
    right_only: list[AnyStr]
    left_list: list[AnyStr]
    right_list: list[AnyStr]
    def report(self) -> None: ...
    def report_partial_closure(self) -> None: ...
    def report_full_closure(self) -> None: ...
    methodmap: dict[str, Callable[[], None]]
    def phase0(self) -> None: ...
    def phase1(self) -> None: ...
    def phase2(self) -> None: ...
    def phase3(self) -> None: ...
    def phase4(self) -> None: ...
    def phase4_closure(self) -> None: ...



from pytype.pytd import pep484
from pytype.pytd import pytd
from pytype.pytd import pytd_utils
from pytype.pytd.parse import parser_test_base
import unittest


class TestPEP484(parser_test_base.ParserTest):
  """Test the visitors in optimize.py."""

  def convert(self, t):
    """Run ConvertTypingToNative and return the result as a string."""
    return pytd_utils.Print(t.Visit(pep484.ConvertTypingToNative(None)))

  def test_convert_optional(self):
    t = pytd.GenericType(pytd.NamedType("typing.Optional"),
                         (pytd.NamedType("str"),))
    self.assertEqual(self.convert(t), "Optional[str]")

  def test_convert_union(self):
    t = pytd.GenericType(pytd.NamedType("typing.Union"),
                         (pytd.NamedType("str"), pytd.NamedType("float")))
    self.assertEqual(self.convert(t), "Union[str, float]")

  def test_convert_list(self):
    t = pytd.NamedType("typing.List")
    self.assertEqual(self.convert(t), "list")

  def test_convert_tuple(self):
    t = pytd.NamedType("typing.Tuple")
    self.assertEqual(self.convert(t), "tuple")

  def test_convert_any(self):
    t = pytd.NamedType("typing.Any")
    self.assertEqual(self.convert(t), "Any")

  def test_convert_anystr(self):
    t = pytd.NamedType("typing.AnyStr")
    self.assertEqual(self.convert(t), "AnyStr")


if __name__ == "__main__":
  unittest.main()



from contextlib import AbstractContextManager
from io import TextIOWrapper
from typing import Any

WINDOWS: bool

def pty_size() -> tuple[int, int]: ...
def stdin_is_foregrounded_tty(stream: Any) -> bool: ...
def cbreak_already_set(stream: TextIOWrapper) -> bool: ...
def character_buffered(stream: TextIOWrapper) -> AbstractContextManager[None]: ...
def ready_for_reading(input_: TextIOWrapper) -> bool: ...
def bytes_to_read(input_: TextIOWrapper) -> int: ...



from typing import Any, Callable, NamedTuple, Text

class Event(NamedTuple):
    time: float
    priority: Any
    action: Callable[..., Any]
    argument: tuple[Any, ...]
    kwargs: dict[Text, Any]

class scheduler:
    def __init__(self, timefunc: Callable[[], float], delayfunc: Callable[[float], None]) -> None: ...
    def enterabs(self, time: float, priority: Any, action: Callable[..., Any], argument: tuple[Any, ...]) -> Event: ...
    def enter(self, delay: float, priority: Any, action: Callable[..., Any], argument: tuple[Any, ...]) -> Event: ...
    def run(self) -> None: ...
    def cancel(self, event: Event) -> None: ...
    def empty(self) -> bool: ...
    @property
    def queue(self) -> list[Event]: ...



from typing import Any

def temporary_chdir(new_dir) -> None: ...
def get_file_directory(): ...
def get_temp_directory(): ...
def get_themes_directory(theme_name: Any | None = ..., png: bool = ...): ...
def create_directory(directory): ...



from .cluster import RedisClusterCommands as RedisClusterCommands
from .core import CoreCommands as CoreCommands
from .helpers import list_or_args as list_or_args
from .parser import CommandsParser as CommandsParser
from .redismodules import RedisModuleCommands as RedisModuleCommands
from .sentinel import SentinelCommands as SentinelCommands

__all__ = ["RedisClusterCommands", "CommandsParser", "CoreCommands", "list_or_args", "RedisModuleCommands", "SentinelCommands"]



from logging import Logger

from aws_xray_sdk.core.daemon_config import DaemonConfig as DaemonConfig

from ..exceptions.exceptions import InvalidDaemonAddressException as InvalidDaemonAddressException

log: Logger
PROTOCOL_HEADER: str
PROTOCOL_DELIMITER: str
DEFAULT_DAEMON_ADDRESS: str

class UDPEmitter:
    def __init__(self, daemon_address=...) -> None: ...
    def send_entity(self, entity) -> None: ...
    def set_daemon_address(self, address) -> None: ...
    @property
    def ip(self): ...
    @property
    def port(self): ...



"""Constructs related to type annotations."""

import collections
import logging
import typing
from typing import Mapping, Tuple, Type, Union as _Union

from pytype import datatypes
from pytype.abstract import _base
from pytype.abstract import _classes
from pytype.abstract import _instance_base
from pytype.abstract import abstract_utils
from pytype.abstract import function
from pytype.abstract import mixin
from pytype.pytd import pytd_utils

log = logging.getLogger(__name__)


class AnnotationClass(_instance_base.SimpleValue, mixin.HasSlots):
  """Base class of annotations that can be parameterized."""

  def __init__(self, name, ctx):
    super().__init__(name, ctx)
    mixin.HasSlots.init_mixin(self)
    self.set_slot("__getitem__", self.getitem_slot)

  def getitem_slot(self, node, slice_var):
    """Custom __getitem__ implementation."""
    slice_content = abstract_utils.maybe_extract_tuple(slice_var)
    inner, ellipses = self._build_inner(slice_content)
    value = self._build_value(node, tuple(inner), ellipses)
    return node, value.to_variable(node)

  def _build_inner(self, slice_content):
    """Build the list of parameters.

    Args:
      slice_content: The iterable of variables to extract parameters from.

    Returns:
      A tuple of a list of parameters and a set of indices at which an ellipsis
        was replaced with Any.
    """
    inner = []
    ellipses = set()
    for var in slice_content:
      if len(var.bindings) > 1:
        self.ctx.errorlog.ambiguous_annotation(self.ctx.vm.frames, var.data)
        inner.append(self.ctx.convert.unsolvable)
      else:
        val = var.bindings[0].data
        if val is self.ctx.convert.ellipsis:
          # Ellipses are allowed only in special cases, so turn them into Any
          # but record the indices so we can check if they're legal.
          ellipses.add(len(inner))
          inner.append(self.ctx.convert.unsolvable)
        else:
          inner.append(val)
    return inner, ellipses

  def _build_value(self, node, inner, ellipses):
    raise NotImplementedError(self.__class__.__name__)

  def __repr__(self):
    return f"AnnotationClass({self.name})"

  def _get_class(self):
    return self.ctx.convert.type_type


class AnnotationContainer(AnnotationClass):
  """Implementation of X[...] for annotations."""

  def __init__(self, name, ctx, base_cls):
    super().__init__(name, ctx)
    self.base_cls = base_cls

  def __repr__(self):
    return f"AnnotationContainer({self.name})"

  def _sub_annotation(
      self, annot: _base.BaseValue, subst: Mapping[str, _base.BaseValue]
  ) -> _base.BaseValue:
    """Apply type parameter substitutions to an annotation."""
    # This is very similar to annotation_utils.sub_one_annotation, but a couple
    # differences make it more convenient to maintain two separate methods:
    # - subst here is a str->BaseValue mapping rather than str->Variable, and it
    #   would be wasteful to create variables just to match sub_one_annotation's
    #   expected input type.
    # - subst contains the type to be substituted in, not an instance of it.
    #   Again, instantiating the type just to later get the type of the instance
    #   is unnecessary extra work.
    if isinstance(annot, TypeParameter):
      if annot.full_name in subst:
        return subst[annot.full_name]
      else:
        return self.ctx.convert.unsolvable
    elif isinstance(annot, mixin.NestedAnnotation):
      inner_types = [(key, self._sub_annotation(val, subst))
                     for key, val in annot.get_inner_types()]
      return annot.replace(inner_types)
    return annot

  def _get_value_info(
      self, inner, ellipses, allowed_ellipses=frozenset()
  ) -> Tuple[Tuple[_Union[int, str], ...], Tuple[_base.BaseValue, ...],
             Type[_classes.ParameterizedClass]]:
    """Get information about the container's inner values.

    Args:
      inner: The list of parameters from _build_inner().
      ellipses: The set of ellipsis indices from _build_inner().
      allowed_ellipses: Optionally, a set of indices at which ellipses are
        allowed. If omitted, ellipses are assumed to be never allowed.

    Returns:
      A tuple of the template, the parameters, and the container class.
    """
    if self.base_cls.full_name == "typing.Protocol":
      return abstract_utils.build_generic_template(inner, self) + (
          _classes.ParameterizedClass,)  # pytype: disable=bad-return-type
    if isinstance(self.base_cls, _classes.TupleClass):
      template = tuple(range(self.base_cls.tuple_length))
    elif isinstance(self.base_cls, _classes.CallableClass):
      template = tuple(range(self.base_cls.num_args)) + (abstract_utils.RET,)
    else:
      template = tuple(t.name for t in self.base_cls.template)
    self.ctx.errorlog.invalid_ellipses(self.ctx.vm.frames,
                                       ellipses - allowed_ellipses, self.name)
    last_index = len(inner) - 1
    if last_index and last_index in ellipses and len(inner) > len(template):
      # Even if an ellipsis is not allowed at this position, strip it off so
      # that we report only one error for something like 'List[int, ...]'
      inner = inner[:-1]
    if isinstance(self.base_cls, _classes.ParameterizedClass):
      # We're dealing with a generic type alias, e.g.:
      #   X = Dict[T, str]
      #   def f(x: X[int]): ...
      # We construct `inner` using both the new inner values and the ones
      # already in X, to end up with a final result of:
      #   template=(_K, _V)
      #   inner=(int, str)
      new_inner = []
      inner_idx = 0
      subst = {}
      # Note that we ignore any missing or extra values in inner for now; the
      # problem will be reported later by _validate_inner.
      for k in template:
        v = self.base_cls.formal_type_parameters[k]
        if v.formal:
          params = self.ctx.annotation_utils.get_type_parameters(v)
          for param in params:
            # If there are too few parameters, we ignore the problem for now;
            # it'll be reported when _build_value checks that the lengths of
            # template and inner match.
            if param.full_name not in subst and inner_idx < len(inner):
              subst[param.full_name] = inner[inner_idx]
              inner_idx += 1
          new_inner.append(self._sub_annotation(v, subst))
        else:
          new_inner.append(v)
      inner = tuple(new_inner)
      if isinstance(self.base_cls, _classes.TupleClass):
        template += (abstract_utils.T,)
        inner += (self.ctx.convert.merge_values(inner),)
      elif isinstance(self.base_cls, _classes.CallableClass):
        template = template[:-1] + (abstract_utils.ARGS,) + template[-1:]
        args = inner[:-1]
        inner = args + (self.ctx.convert.merge_values(args),) + inner[-1:]
      abstract_class = type(self.base_cls)
    else:
      abstract_class = _classes.ParameterizedClass
    return template, inner, abstract_class

  def _validate_inner(self, template, inner, raw_inner):
    """Check that the passed inner values are valid for the given template."""
    if (isinstance(self.base_cls, _classes.ParameterizedClass) and
        not abstract_utils.is_generic_protocol(self.base_cls)):
      # For a generic type alias, we check that the number of typevars in the
      # alias matches the number of raw parameters provided.
      template_length = raw_template_length = len(
          set(self.ctx.annotation_utils.get_type_parameters(self.base_cls)))
      inner_length = len(raw_inner)
      base_cls = self.base_cls.base_cls
    else:
      # In all other cases, we check that the final template length and
      # parameter count match, after any adjustments like flattening the inner
      # argument list in a Callable.
      template_length = len(template)
      raw_template_length = len(self.base_cls.template)
      inner_length = len(inner)
      base_cls = self.base_cls
    if inner_length != template_length:
      if not template:
        self.ctx.errorlog.not_indexable(
            self.ctx.vm.frames, base_cls.name, generic_warning=True)
      else:
        # Use the unprocessed values of `template` and `inner` so that the error
        # message matches what the user sees.
        if isinstance(self.base_cls, _classes.ParameterizedClass):
          error_template = None
        else:
          error_template = (t.name for t in base_cls.template)
        self.ctx.errorlog.wrong_annotation_parameter_count(
            self.ctx.vm.frames, self.base_cls, raw_inner, raw_template_length,
            error_template)
    else:
      if len(inner) == 1:
        val, = inner
        # It's a common mistake to index a container class rather than an
        # instance (e.g., list[0]).
        # We only check the "int" case, since string literals are allowed for
        # late annotations.
        if (isinstance(val, _instance_base.Instance) and
            val.cls == self.ctx.convert.int_type):
          # Don't report this error again.
          inner = (self.ctx.convert.unsolvable,)
          self.ctx.errorlog.not_indexable(self.ctx.vm.frames, self.name)
      # Check for a misused Final annotation
      if any(isinstance(val, FinalAnnotation) for val in inner):
        self.ctx.errorlog.invalid_final_type(self.ctx.vm.frames)
        inner = [val.annotation if isinstance(val, FinalAnnotation) else val
                 for val in inner]
    return inner

  def _build_value(self, node, inner, ellipses):
    if self.base_cls.is_late_annotation():
      # A parameterized LateAnnotation should be converted to another
      # LateAnnotation to delay evaluation until the first late annotation is
      # resolved. We don't want to create a ParameterizedClass immediately
      # because (1) ParameterizedClass expects its base_cls to be a
      # class_mixin.Class, and (2) we have to postpone error-checking anyway so
      # we might as well postpone the entire evaluation.
      printed_params = []
      added_imports = collections.defaultdict(set)
      for i, param in enumerate(inner):
        if i in ellipses:
          printed_params.append("...")
        else:
          typ = param.get_instance_type(node)
          annot, imports = pytd_utils.MakeTypeAnnotation(typ)
          printed_params.append(annot)
          for k, v in imports.items():
            added_imports[k] |= v

      expr = f"{self.base_cls.expr}[{', '.join(printed_params)}]"
      annot = LateAnnotation(expr, self.base_cls.stack, self.ctx,
                             imports=added_imports)
      self.ctx.vm.late_annotations[self.base_cls.expr].append(annot)
      return annot
    template, processed_inner, abstract_class = self._get_value_info(
        inner, ellipses)
    if isinstance(self.base_cls, _classes.ParameterizedClass):
      base_cls = self.base_cls.base_cls
    else:
      base_cls = self.base_cls
    if base_cls.full_name in ("typing.Generic", "typing.Protocol"):
      # Generic is unique in that parameterizing it defines a new template;
      # usually, the parameterized class inherits the base class's template.
      # Protocol[T, ...] is a shorthand for Protocol, Generic[T, ...].
      template_params = [
          param.with_module(base_cls.full_name)
          for param in typing.cast(Tuple[TypeParameter, ...], processed_inner)]
    else:
      template_params = None
    processed_inner = self._validate_inner(template, processed_inner, inner)
    params = {
        name: (processed_inner[i]
               if i < len(processed_inner) else self.ctx.convert.unsolvable)
        for i, name in enumerate(template)
    }

    # For user-defined generic types, check if its type parameter matches
    # its corresponding concrete type
    if isinstance(base_cls, _classes.InterpreterClass) and base_cls.template:
      for formal_param in base_cls.template:
        root_node = self.ctx.root_node
        param_value = params[formal_param.name]
        if (isinstance(formal_param, TypeParameter) and
            not formal_param.is_generic() and
            isinstance(param_value, TypeParameter)):
          if formal_param.name == param_value.name:
            # We don't need to check if a TypeParameter matches itself.
            continue
          else:
            actual = param_value.instantiate(
                root_node, container=abstract_utils.DUMMY_CONTAINER)
        else:
          actual = param_value.instantiate(root_node)
        bad, _ = self.ctx.matcher(root_node).bad_matches(actual, formal_param)
        if bad:
          if not isinstance(param_value, TypeParameter):
            # If param_value is not a TypeVar, we substitute in TypeVar bounds
            # and constraints in formal_param for a more helpful error message.
            formal_param = self.ctx.annotation_utils.sub_one_annotation(
                root_node, formal_param, [{}])
            details = None
          elif isinstance(formal_param, TypeParameter):
            details = (f"TypeVars {formal_param.name} and {param_value.name} "
                       "have incompatible bounds or constraints.")
          else:
            details = None
          self.ctx.errorlog.bad_concrete_type(
              self.ctx.vm.frames, root_node, formal_param, actual, bad, details)
          return self.ctx.convert.unsolvable

    try:
      return abstract_class(base_cls, params, self.ctx, template_params)
    except abstract_utils.GenericTypeError as e:
      self.ctx.errorlog.invalid_annotation(self.ctx.vm.frames, e.annot, e.error)
      return self.ctx.convert.unsolvable

  def call(self, node, func, args, alias_map=None):
    return self._call_helper(node, self.base_cls, func, args)


class TypeParameter(_base.BaseValue):
  """Parameter of a type."""

  formal = True

  def __init__(self,
               name,
               ctx,
               constraints=(),
               bound=None,
               covariant=False,
               contravariant=False,
               module=None):
    super().__init__(name, ctx)
    self.constraints = constraints
    self.bound = bound
    self.covariant = covariant
    self.contravariant = contravariant
    self.module = module

  def is_generic(self):
    return not self.constraints and not self.bound

  def copy(self):
    return TypeParameter(self.name, self.ctx, self.constraints, self.bound,
                         self.covariant, self.contravariant, self.module)

  def with_module(self, module):
    res = self.copy()
    res.module = module
    return res

  def __eq__(self, other):
    if isinstance(other, type(self)):
      return (self.name == other.name and
              self.constraints == other.constraints and
              self.bound == other.bound and
              self.covariant == other.covariant and
              self.contravariant == other.contravariant and
              self.module == other.module)
    return NotImplemented

  def __ne__(self, other):
    return not self == other

  def __hash__(self):
    return hash((self.name, self.constraints, self.bound, self.covariant,
                 self.contravariant))

  def __repr__(self):
    return ("TypeParameter({!r}, constraints={!r}, bound={!r}, module={!r})"
            .format(self.name, self.constraints, self.bound, self.module))

  def instantiate(self, node, container=None):
    var = self.ctx.program.NewVariable()
    if container and (not isinstance(container, _instance_base.SimpleValue) or
                      self.full_name in container.all_template_names):
      instance = TypeParameterInstance(self, container, self.ctx)
      return instance.to_variable(node)
    else:
      for c in self.constraints:
        var.PasteVariable(c.instantiate(node, container))
      if self.bound:
        var.PasteVariable(self.bound.instantiate(node, container))
    if not var.bindings:
      var.AddBinding(self.ctx.convert.unsolvable, [], node)
    return var

  def update_official_name(self, name):
    if self.name != name:
      message = (f"TypeVar({self.name!r}) must be stored as {self.name!r}, "
                 f"not {name!r}")
      self.ctx.errorlog.invalid_typevar(self.ctx.vm.frames, message)

  def call(self, node, func, args, alias_map=None):
    return node, self.instantiate(node)


class TypeParameterInstance(_base.BaseValue):
  """An instance of a type parameter."""

  def __init__(self, param, instance, ctx):
    super().__init__(param.name, ctx)
    self.cls = self.param = param
    self.instance = instance
    self.module = param.module

  def call(self, node, func, args, alias_map=None):
    var = self.instance.get_instance_type_parameter(self.name)
    if var.bindings:
      return function.call_function(self.ctx, node, var, args)
    else:
      return node, self.ctx.convert.empty.to_variable(self.ctx.root_node)

  def __repr__(self):
    return f"TypeParameterInstance({self.name!r})"

  def __eq__(self, other):
    if isinstance(other, type(self)):
      return self.param == other.param and self.instance == other.instance
    return NotImplemented

  def __hash__(self):
    return hash((self.param, self.instance))


class Union(_base.BaseValue, mixin.NestedAnnotation, mixin.HasSlots):
  """A list of types.

  Used for parameter matching.

  Attributes:
    options: Iterable of instances of BaseValue.
  """

  def __init__(self, options, ctx):
    super().__init__("Union", ctx)
    assert options
    self.options = list(options)
    self.cls = self._get_class()
    self._printing = False
    self._instance_cache = {}
    mixin.NestedAnnotation.init_mixin(self)
    mixin.HasSlots.init_mixin(self)
    self.set_slot("__getitem__", self.getitem_slot)

  def __repr__(self):
    if self._printing:  # recursion detected
      printed_contents = "..."
    else:
      self._printing = True
      printed_contents = ", ".join(repr(o) for o in self.options)
      self._printing = False
    return f"{self.name}[{printed_contents}]"

  def __eq__(self, other):
    if isinstance(other, type(self)):
      return self.options == other.options
    return NotImplemented

  def __ne__(self, other):
    return not self == other

  def __hash__(self):
    # Use the names of the parameter values to approximate a hash, to avoid
    # infinite recursion on recursive type annotations.
    return hash(tuple(o.full_name for o in self.options))

  def _unique_parameters(self):
    return [o.to_variable(self.ctx.root_node) for o in self.options]

  def _get_class(self):
    classes = {o.cls for o in self.options}
    if len(classes) > 1:
      return self.ctx.convert.unsolvable
    else:
      return classes.pop()

  def getitem_slot(self, node, slice_var):
    """Custom __getitem__ implementation."""
    slice_content = abstract_utils.maybe_extract_tuple(slice_var)
    params = self.ctx.annotation_utils.get_type_parameters(self)
    num_params = len({x.name for x in params})
    # Check that we are instantiating all the unbound type parameters
    if num_params != len(slice_content):
      self.ctx.errorlog.wrong_annotation_parameter_count(
          self.ctx.vm.frames, self, [v.data[0] for v in slice_content],
          num_params)
      return node, self.ctx.new_unsolvable(node)
    concrete = (
        var.data[0].instantiate(node, container=abstract_utils.DUMMY_CONTAINER)
        for var in slice_content)
    subst = datatypes.AliasingDict()
    for p in params:
      for k in subst:
        if k == p.name or k.endswith(f".{p.name}"):
          subst.add_alias(p.full_name, k)
          break
      else:
        subst[p.full_name] = next(concrete)
    new = self.ctx.annotation_utils.sub_one_annotation(node, self, [subst])
    return node, new.to_variable(node)

  def instantiate(self, node, container=None):
    var = self.ctx.program.NewVariable()
    for option in self.options:
      k = (node, container, option)
      if k in self._instance_cache:
        if self._instance_cache[k] is None:
          self._instance_cache[k] = self.ctx.new_unsolvable(node)
        instance = self._instance_cache[k]
      else:
        self._instance_cache[k] = None
        instance = option.instantiate(node, container)
        self._instance_cache[k] = instance
      var.PasteVariable(instance, node)
    return var

  def call(self, node, func, args, alias_map=None):
    var = self.ctx.program.NewVariable(self.options, [], node)
    return function.call_function(self.ctx, node, var, args)

  def get_formal_type_parameter(self, t):
    new_options = [option.get_formal_type_parameter(t)
                   for option in self.options]
    return Union(new_options, self.ctx)

  def get_inner_types(self):
    return enumerate(self.options)

  def update_inner_type(self, key, typ):
    self.options[key] = typ

  def replace(self, inner_types):
    return self.__class__((v for _, v in sorted(inner_types)), self.ctx)


class LateAnnotation:
  """A late annotation.

  A late annotation stores a string expression and a snapshot of the VM stack at
  the point where the annotation was introduced. Once the expression is
  resolved, the annotation pretends to be the resolved type; before that, it
  pretends to be an unsolvable. This effect is achieved by delegating attribute
  lookup with __getattribute__.

  Note that for late annotation x, `isinstance(x, ...)` and `x.__class__` will
  use the type that x is pretending to be; `type(x)` will reveal x's true type.
  Use `x.is_late_annotation()` to check whether x is a late annotation.
  """

  _RESOLVING = object()

  def __init__(self, expr, stack, ctx, *, imports=None):
    self.expr = expr
    self.stack = stack
    self.ctx = ctx
    self.resolved = False
    # Any new imports the annotation needs while resolving.
    self._imports = imports or {}
    self._type = ctx.convert.unsolvable  # the resolved type of `expr`
    self._unresolved_instances = set()
    # _attribute_names needs to be defined last! This contains the names of all
    # of LateAnnotation's attributes, discovered by looking at
    # LateAnnotation.__dict__ and self.__dict__. These names are used in
    # __getattribute__ and __setattr__ to determine whether a given get/setattr
    # call should operate on the LateAnnotation itself or its resolved type.
    self._attribute_names = (
        set(LateAnnotation.__dict__) |
        set(super().__getattribute__("__dict__")))

  def flatten_expr(self):
    """Flattens the expression into a legal variable name if necessary.

    Pytype stores parameterized recursive types in intermediate variables. If
    self is such a type, this method flattens self.expr into a string that can
    serve as a variable name. For example, 'MyRecursiveAlias[int, str]' is
    flattened into '_MyRecursiveAlias_LBAR_int_COMMA_str_RBAR'.

    Returns:
      If self is a parameterized recursive type, a flattened version of
      self.expr that is a legal variable name. Otherwise, self.expr unchanged.
    """
    if "[" in self.expr and self.is_recursive():
      # _DOT and _RBAR have no trailing underscore because they precede names
      # that we already prefix an underscore to.
      return "_" + self.expr.replace(".", "_DOT").replace(
          "[", "_LBAR_").replace("]", "_RBAR").replace(", ", "_COMMA_")
    return self.expr

  def unflatten_expr(self):
    """Unflattens a flattened expression."""
    if "_LBAR_" in self.expr:
      mod, dot, rest = self.expr.rpartition(".")
      # The [1:] slicing and trailing underscore in _DOT_ are to get rid of
      # leading underscores added when flattening.
      return mod + dot + rest[1:].replace("_DOT_", ".").replace(
          "_LBAR_", "[").replace("_RBAR", "]").replace("_COMMA_", ", ")
    return self.expr

  def __repr__(self):
    return "LateAnnotation({!r}, resolved={!r})".format(
        self.expr, self._type if self.resolved else None)

  # __hash__ and __eq__ need to be explicitly defined for Python to use them in
  # set/dict comparisons.

  def __hash__(self):
    return hash(self._type) if self.resolved else hash(self.expr)

  def __eq__(self, other):
    return hash(self) == hash(other)

  def __getattribute__(self, name):
    if name == "_attribute_names" or name in self._attribute_names:
      return super().__getattribute__(name)
    return self._type.__getattribute__(name)  # pytype: disable=attribute-error

  def __setattr__(self, name, value):
    if not hasattr(self, "_attribute_names") or name in self._attribute_names:
      return super().__setattr__(name, value)
    return self._type.__setattr__(name, value)

  def resolve(self, node, f_globals, f_locals):
    """Resolve the late annotation."""
    if self.resolved:
      return
    # Sets resolved to a truthy value distinguishable from True so that
    # 'if self.resolved' is True when self is partially resolved, but code that
    # really needs to tell partially and fully resolved apart can do so.
    self.resolved = LateAnnotation._RESOLVING
    # Add implicit imports for typing, since we can have late annotations like
    # `set[int]` which get converted to `typing.Set[int]`.
    if "typing" in self._imports:
      mod = self.ctx.loader.import_name("typing")
      for v in self._imports["typing"]:
        if v not in f_globals.members:
          if v == "Any":
            typ = self.ctx.convert.unsolvable
          else:
            typ = self.ctx.convert.name_to_value(f"typing.{v}", ast=mod)
          f_globals.members[v] = typ.to_variable(node)
    var, errorlog = abstract_utils.eval_expr(self.ctx, node, f_globals,
                                             f_locals, self.expr)
    if errorlog:
      self.ctx.errorlog.copy_from(errorlog.errors, self.stack)
    self._type = self.ctx.annotation_utils.extract_annotation(
        node, var, None, self.stack)
    if self._type != self.ctx.convert.unsolvable:
      # We may have tried to call __init__ on instances of this annotation.
      # Since the annotation was unresolved at the time, we need to call
      # __init__ again to define any instance attributes.
      for instance in self._unresolved_instances:
        if isinstance(instance.cls, Union):
          # Having instance.cls be a Union type will crash in attribute.py.
          # Setting it to Any picks up the annotation in another code path.
          instance.cls = self.ctx.convert.unsolvable
        else:
          self.ctx.vm.reinitialize_if_initialized(node, instance)
    log.info("Resolved late annotation %r to %r", self.expr, self._type)

  def set_type(self, typ):
    # Used by annotation_utils.sub_one_annotation to substitute values into
    # recursive aliases.
    assert not self.resolved
    self.resolved = True
    self._type = typ

  def to_variable(self, node):
    if self.resolved:
      return self._type.to_variable(node)
    else:
      return _base.BaseValue.to_variable(self, node)  # pytype: disable=wrong-arg-types

  def instantiate(self, node, container=None):
    """Instantiate the pointed-to class, or record a placeholder instance."""
    if self.resolved:
      return self._type.instantiate(node, container)
    else:
      instance = _instance_base.Instance(self, self.ctx)
      self._unresolved_instances.add(instance)
      return instance.to_variable(node)

  def get_special_attribute(self, node, name, valself):
    if name == "__getitem__" and not self.resolved:
      container = _base.BaseValue.to_annotation_container(self)  # pytype: disable=wrong-arg-types
      return container.get_special_attribute(node, name, valself)
    return self._type.get_special_attribute(node, name, valself)

  def is_late_annotation(self):
    return True

  def is_recursive(self):
    """Check whether this is a recursive type."""
    if not self.resolved:
      return False
    seen = {id(self)}
    stack = [self._type]
    while stack:
      t = stack.pop()
      if t.is_late_annotation():
        if id(t) in seen:
          return True
        seen.add(id(t))
      if isinstance(t, mixin.NestedAnnotation):
        stack.extend(child for _, child in t.get_inner_types())
    return False


class FinalAnnotation(_base.BaseValue):
  """Container for a Final annotation."""

  def __init__(self, annotation, ctx):
    super().__init__("FinalAnnotation", ctx)
    self.annotation = annotation

  def __repr__(self):
    return f"Final[{self.annotation}]"

  def instantiate(self, node, container=None):
    return self.to_variable(node)



import sys
from typing import (
    Any,
    AnyStr,
    Callable,
    Container,
    Generic,
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Protocol,
    Sequence,
    SupportsAbs,
    TypeVar,
    overload,
)
from typing_extensions import ParamSpec, SupportsIndex, final

_R = TypeVar("_R")
_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_K = TypeVar("_K")
_V = TypeVar("_V")
_P = ParamSpec("_P")

# The following protocols return "Any" instead of bool, since the comparison
# operators can be overloaded to return an arbitrary object. For example,
# the numpy.array comparison dunders return another numpy.array.

class _SupportsDunderLT(Protocol):
    def __lt__(self, __other: Any) -> Any: ...

class _SupportsDunderGT(Protocol):
    def __gt__(self, __other: Any) -> Any: ...

class _SupportsDunderLE(Protocol):
    def __le__(self, __other: Any) -> Any: ...

class _SupportsDunderGE(Protocol):
    def __ge__(self, __other: Any) -> Any: ...

_SupportsComparison = _SupportsDunderLE | _SupportsDunderGE | _SupportsDunderGT | _SupportsDunderLT

class _SupportsInversion(Protocol[_T_co]):
    def __invert__(self) -> _T_co: ...

class _SupportsNeg(Protocol[_T_co]):
    def __neg__(self) -> _T_co: ...

class _SupportsPos(Protocol[_T_co]):
    def __pos__(self) -> _T_co: ...

# All four comparison functions must have the same signature, or we get false-positive errors
def lt(__a: _SupportsComparison, __b: _SupportsComparison) -> Any: ...
def le(__a: _SupportsComparison, __b: _SupportsComparison) -> Any: ...
def eq(__a: object, __b: object) -> Any: ...
def ne(__a: object, __b: object) -> Any: ...
def ge(__a: _SupportsComparison, __b: _SupportsComparison) -> Any: ...
def gt(__a: _SupportsComparison, __b: _SupportsComparison) -> Any: ...
def not_(__a: object) -> bool: ...
def truth(__a: object) -> bool: ...
def is_(__a: object, __b: object) -> bool: ...
def is_not(__a: object, __b: object) -> bool: ...
def abs(__a: SupportsAbs[_T]) -> _T: ...
def add(__a: Any, __b: Any) -> Any: ...
def and_(__a: Any, __b: Any) -> Any: ...
def floordiv(__a: Any, __b: Any) -> Any: ...
def index(__a: SupportsIndex) -> int: ...
def inv(__a: _SupportsInversion[_T_co]) -> _T_co: ...
def invert(__a: _SupportsInversion[_T_co]) -> _T_co: ...
def lshift(__a: Any, __b: Any) -> Any: ...
def mod(__a: Any, __b: Any) -> Any: ...
def mul(__a: Any, __b: Any) -> Any: ...
def matmul(__a: Any, __b: Any) -> Any: ...
def neg(__a: _SupportsNeg[_T_co]) -> _T_co: ...
def or_(__a: Any, __b: Any) -> Any: ...
def pos(__a: _SupportsPos[_T_co]) -> _T_co: ...
def pow(__a: Any, __b: Any) -> Any: ...
def rshift(__a: Any, __b: Any) -> Any: ...
def sub(__a: Any, __b: Any) -> Any: ...
def truediv(__a: Any, __b: Any) -> Any: ...
def xor(__a: Any, __b: Any) -> Any: ...
def concat(__a: Sequence[_T], __b: Sequence[_T]) -> Sequence[_T]: ...
def contains(__a: Container[object], __b: object) -> bool: ...
def countOf(__a: Iterable[object], __b: object) -> int: ...
@overload
def delitem(__a: MutableSequence[Any], __b: SupportsIndex) -> None: ...
@overload
def delitem(__a: MutableSequence[Any], __b: slice) -> None: ...
@overload
def delitem(__a: MutableMapping[_K, Any], __b: _K) -> None: ...
@overload
def getitem(__a: Sequence[_T], __b: SupportsIndex) -> _T: ...
@overload
def getitem(__a: Sequence[_T], __b: slice) -> Sequence[_T]: ...
@overload
def getitem(__a: Mapping[_K, _V], __b: _K) -> _V: ...
def indexOf(__a: Iterable[_T], __b: _T) -> int: ...
@overload
def setitem(__a: MutableSequence[_T], __b: SupportsIndex, __c: _T) -> None: ...
@overload
def setitem(__a: MutableSequence[_T], __b: slice, __c: Sequence[_T]) -> None: ...
@overload
def setitem(__a: MutableMapping[_K, _V], __b: _K, __c: _V) -> None: ...
def length_hint(__obj: object, __default: int = ...) -> int: ...
@final
class attrgetter(Generic[_T_co]):
    @overload
    def __new__(cls, attr: str) -> attrgetter[Any]: ...
    @overload
    def __new__(cls, attr: str, __attr2: str) -> attrgetter[tuple[Any, Any]]: ...
    @overload
    def __new__(cls, attr: str, __attr2: str, __attr3: str) -> attrgetter[tuple[Any, Any, Any]]: ...
    @overload
    def __new__(cls, attr: str, __attr2: str, __attr3: str, __attr4: str) -> attrgetter[tuple[Any, Any, Any, Any]]: ...
    @overload
    def __new__(cls, attr: str, *attrs: str) -> attrgetter[tuple[Any, ...]]: ...
    def __call__(self, obj: Any) -> _T_co: ...

@final
class itemgetter(Generic[_T_co]):
    @overload
    def __new__(cls, item: Any) -> itemgetter[Any]: ...
    @overload
    def __new__(cls, item: Any, __item2: Any) -> itemgetter[tuple[Any, Any]]: ...
    @overload
    def __new__(cls, item: Any, __item2: Any, __item3: Any) -> itemgetter[tuple[Any, Any, Any]]: ...
    @overload
    def __new__(cls, item: Any, __item2: Any, __item3: Any, __item4: Any) -> itemgetter[tuple[Any, Any, Any, Any]]: ...
    @overload
    def __new__(cls, item: Any, *items: Any) -> itemgetter[tuple[Any, ...]]: ...
    def __call__(self, obj: Any) -> _T_co: ...

@final
class methodcaller:
    def __init__(self, __name: str, *args: Any, **kwargs: Any) -> None: ...
    def __call__(self, obj: Any) -> Any: ...

def iadd(__a: Any, __b: Any) -> Any: ...
def iand(__a: Any, __b: Any) -> Any: ...
def iconcat(__a: Any, __b: Any) -> Any: ...
def ifloordiv(__a: Any, __b: Any) -> Any: ...
def ilshift(__a: Any, __b: Any) -> Any: ...
def imod(__a: Any, __b: Any) -> Any: ...
def imul(__a: Any, __b: Any) -> Any: ...
def imatmul(__a: Any, __b: Any) -> Any: ...
def ior(__a: Any, __b: Any) -> Any: ...
def ipow(__a: Any, __b: Any) -> Any: ...
def irshift(__a: Any, __b: Any) -> Any: ...
def isub(__a: Any, __b: Any) -> Any: ...
def itruediv(__a: Any, __b: Any) -> Any: ...
def ixor(__a: Any, __b: Any) -> Any: ...

if sys.version_info >= (3, 11):
    def call(__obj: Callable[_P, _R], *args: _P.args, **kwargs: _P.kwargs) -> _R: ...

def _compare_digest(__a: AnyStr, __b: AnyStr) -> bool: ...



from . import _LangModelType

Latin2_HungarianCharToOrderMap: tuple[int, ...]
win1250HungarianCharToOrderMap: tuple[int, ...]
HungarianLangModel: tuple[int, ...]
Latin2HungarianModel: _LangModelType
Win1250HungarianModel: _LangModelType



import sys
from tkinter.commondialog import Dialog
from typing import Any, ClassVar

if sys.version_info >= (3, 9):
    __all__ = [
        "showinfo",
        "showwarning",
        "showerror",
        "askquestion",
        "askokcancel",
        "askyesno",
        "askyesnocancel",
        "askretrycancel",
    ]

ERROR: str
INFO: str
QUESTION: str
WARNING: str
ABORTRETRYIGNORE: str
OK: str
OKCANCEL: str
RETRYCANCEL: str
YESNO: str
YESNOCANCEL: str
ABORT: str
RETRY: str
IGNORE: str
CANCEL: str
YES: str
NO: str

class Message(Dialog):
    command: ClassVar[str]

def showinfo(title: str | None = ..., message: str | None = ..., **options: Any) -> str: ...
def showwarning(title: str | None = ..., message: str | None = ..., **options: Any) -> str: ...
def showerror(title: str | None = ..., message: str | None = ..., **options: Any) -> str: ...
def askquestion(title: str | None = ..., message: str | None = ..., **options: Any) -> str: ...
def askokcancel(title: str | None = ..., message: str | None = ..., **options: Any) -> bool: ...
def askyesno(title: str | None = ..., message: str | None = ..., **options: Any) -> bool: ...
def askyesnocancel(title: str | None = ..., message: str | None = ..., **options: Any) -> bool | None: ...
def askretrycancel(title: str | None = ..., message: str | None = ..., **options: Any) -> bool: ...



from typing import MutableSequence, Text

def reset() -> None: ...
def listdir(path: Text) -> list[str]: ...

opendir = listdir

def annotate(head: Text, list: MutableSequence[str] | MutableSequence[Text] | MutableSequence[str | Text]) -> None: ...



import sys
from typing import Any, NamedTuple
from typing_extensions import final

_TimeTuple = tuple[int, int, int, int, int, int, int, int, int]

accept2dyear: bool
altzone: int
daylight: int
timezone: int
tzname: tuple[str, str]

class _struct_time(NamedTuple):
    tm_year: int
    tm_mon: int
    tm_mday: int
    tm_hour: int
    tm_min: int
    tm_sec: int
    tm_wday: int
    tm_yday: int
    tm_isdst: int
    @property
    def n_fields(self) -> int: ...
    @property
    def n_sequence_fields(self) -> int: ...
    @property
    def n_unnamed_fields(self) -> int: ...

@final
class struct_time(_struct_time):
    def __init__(self, o: _TimeTuple, _arg: Any = ...) -> None: ...
    def __new__(cls, o: _TimeTuple, _arg: Any = ...) -> struct_time: ...

def asctime(t: _TimeTuple | struct_time = ...) -> str: ...
def clock() -> float: ...
def ctime(secs: float | None = ...) -> str: ...
def gmtime(secs: float | None = ...) -> struct_time: ...
def localtime(secs: float | None = ...) -> struct_time: ...
def mktime(t: _TimeTuple | struct_time) -> float: ...
def sleep(secs: float) -> None: ...
def strftime(format: str, t: _TimeTuple | struct_time = ...) -> str: ...
def strptime(string: str, format: str = ...) -> struct_time: ...
def time() -> float: ...

if sys.platform != "win32":
    def tzset() -> None: ...  # Unix only



from typing import Any

from selenium.common.exceptions import (
    NoSuchElementException as NoSuchElementException,
    UnexpectedTagNameException as UnexpectedTagNameException,
)
from selenium.webdriver.common.by import By as By

class Select:
    is_multiple: Any
    def __init__(self, webelement) -> None: ...
    @property
    def options(self): ...
    @property
    def all_selected_options(self): ...
    @property
    def first_selected_option(self): ...
    def select_by_value(self, value) -> None: ...
    def select_by_index(self, index) -> None: ...
    def select_by_visible_text(self, text) -> None: ...
    def deselect_all(self) -> None: ...
    def deselect_by_value(self, value) -> None: ...
    def deselect_by_index(self, index) -> None: ...
    def deselect_by_visible_text(self, text) -> None: ...



from typing import Any

from . import roles
from .base import Executable, SchemaVisitor
from .elements import ClauseElement

class _DDLCompiles(ClauseElement): ...

class DDLElement(roles.DDLRole, Executable, _DDLCompiles):
    target: Any
    on: Any
    dialect: Any
    callable_: Any
    def execute(self, bind: Any | None = ..., target: Any | None = ...): ...  # type: ignore[override]
    def against(self, target) -> None: ...
    state: Any
    def execute_if(self, dialect: Any | None = ..., callable_: Any | None = ..., state: Any | None = ...) -> None: ...
    def __call__(self, target, bind, **kw): ...
    bind: Any

class DDL(DDLElement):
    __visit_name__: str
    statement: Any
    context: Any
    def __init__(self, statement, context: Any | None = ..., bind: Any | None = ...) -> None: ...

class _CreateDropBase(DDLElement):
    element: Any
    bind: Any
    if_exists: Any
    if_not_exists: Any
    def __init__(
        self, element, bind: Any | None = ..., if_exists: bool = ..., if_not_exists: bool = ..., _legacy_bind: Any | None = ...
    ) -> None: ...
    @property
    def stringify_dialect(self): ...

class CreateSchema(_CreateDropBase):
    __visit_name__: str
    quote: Any
    def __init__(self, name, quote: Any | None = ..., **kw) -> None: ...

class DropSchema(_CreateDropBase):
    __visit_name__: str
    quote: Any
    cascade: Any
    def __init__(self, name, quote: Any | None = ..., cascade: bool = ..., **kw) -> None: ...

class CreateTable(_CreateDropBase):
    __visit_name__: str
    columns: Any
    include_foreign_key_constraints: Any
    def __init__(
        self, element, bind: Any | None = ..., include_foreign_key_constraints: Any | None = ..., if_not_exists: bool = ...
    ) -> None: ...

class _DropView(_CreateDropBase):
    __visit_name__: str

class CreateColumn(_DDLCompiles):
    __visit_name__: str
    element: Any
    def __init__(self, element) -> None: ...

class DropTable(_CreateDropBase):
    __visit_name__: str
    def __init__(self, element, bind: Any | None = ..., if_exists: bool = ...) -> None: ...

class CreateSequence(_CreateDropBase):
    __visit_name__: str

class DropSequence(_CreateDropBase):
    __visit_name__: str

class CreateIndex(_CreateDropBase):
    __visit_name__: str
    def __init__(self, element, bind: Any | None = ..., if_not_exists: bool = ...) -> None: ...

class DropIndex(_CreateDropBase):
    __visit_name__: str
    def __init__(self, element, bind: Any | None = ..., if_exists: bool = ...) -> None: ...

class AddConstraint(_CreateDropBase):
    __visit_name__: str
    def __init__(self, element, *args, **kw) -> None: ...

class DropConstraint(_CreateDropBase):
    __visit_name__: str
    cascade: Any
    def __init__(self, element, cascade: bool = ..., **kw) -> None: ...

class SetTableComment(_CreateDropBase):
    __visit_name__: str

class DropTableComment(_CreateDropBase):
    __visit_name__: str

class SetColumnComment(_CreateDropBase):
    __visit_name__: str

class DropColumnComment(_CreateDropBase):
    __visit_name__: str

class DDLBase(SchemaVisitor):
    connection: Any
    def __init__(self, connection) -> None: ...

class SchemaGenerator(DDLBase):
    checkfirst: Any
    tables: Any
    preparer: Any
    dialect: Any
    memo: Any
    def __init__(self, dialect, connection, checkfirst: bool = ..., tables: Any | None = ..., **kwargs) -> None: ...
    def visit_metadata(self, metadata) -> None: ...
    def visit_table(
        self, table, create_ok: bool = ..., include_foreign_key_constraints: Any | None = ..., _is_metadata_operation: bool = ...
    ) -> None: ...
    def visit_foreign_key_constraint(self, constraint) -> None: ...
    def visit_sequence(self, sequence, create_ok: bool = ...) -> None: ...
    def visit_index(self, index, create_ok: bool = ...) -> None: ...

class SchemaDropper(DDLBase):
    checkfirst: Any
    tables: Any
    preparer: Any
    dialect: Any
    memo: Any
    def __init__(self, dialect, connection, checkfirst: bool = ..., tables: Any | None = ..., **kwargs) -> None: ...
    def visit_metadata(self, metadata): ...
    def visit_index(self, index, drop_ok: bool = ...) -> None: ...
    def visit_table(self, table, drop_ok: bool = ..., _is_metadata_operation: bool = ..., _ignore_sequences=...) -> None: ...
    def visit_foreign_key_constraint(self, constraint) -> None: ...
    def visit_sequence(self, sequence, drop_ok: bool = ...) -> None: ...

def sort_tables(tables, skip_fn: Any | None = ..., extra_dependencies: Any | None = ...): ...
def sort_tables_and_constraints(
    tables, filter_fn: Any | None = ..., extra_dependencies: Any | None = ..., _warn_for_cycles: bool = ...
): ...



"""Merge .pyi file annotations into a .py file."""

import argparse
import logging
import sys

from pytype.tools.merge_pyi import merge_pyi


def parse_args(argv):
  """Process command line arguments using argparse."""

  parser = argparse.ArgumentParser(
      description='Populate file.py with type annotations from file.pyi.',
      epilog='Outputs merged file to stdout if no other option is set.')

  def check_verbosity(v):
    v = int(v)  # may raise ValueError
    if -1 <= v <= 4:
      return v
    raise ValueError()
  parser.add_argument(
      '-v', '--verbosity', type=check_verbosity, action='store', default=1,
      help=('Set logging verbosity: '
            '-1=quiet, 0=fatal, 1=error (default), 2=warn, 3=info, 4=debug'))

  group = parser.add_mutually_exclusive_group()

  group.add_argument(
      '-i', '--in-place', action='store_true', help='overwrite file.py')

  group.add_argument('--diff', action='store_true', help='print out a diff')

  parser.add_argument(
      '-b', '--backup', type=str,
      help=('extension to use for a backup file, if --in-place is set. '
            'e.g. `-i -b orig` will copy file.py to file.py.orig if '
            'any changes are made.'))

  parser.add_argument(
      'py', type=str, metavar='file.py',
      help='python file to annotate')

  parser.add_argument(
      'pyi', type=str, metavar='file.pyi',
      help='PEP484 stub file with annotations for file.py')

  args = parser.parse_args(argv[1:])

  # Validate args
  if args.backup and not args.in_place:
    parser.error(
        'Cannot use argument -b/--backup without argument -i/--in-place')

  return args


def main(argv=None):
  """Merge a source file and a pyi file.

  Args:
    argv: Flags and files to process.
  """

  if argv is None:
    argv = sys.argv
  args = parse_args(argv)

  # Log levels range from 10 (DEBUG) to 50 (CRITICAL) in increments of 10. A
  # level >50 prevents anything from being logged.
  logging.basicConfig(level=50-args.verbosity*10)

  if args.diff:
    mode = merge_pyi.Mode.DIFF
  elif args.in_place:
    mode = merge_pyi.Mode.OVERWRITE
  else:
    mode = merge_pyi.Mode.PRINT

  backup = args.backup or None
  changed = merge_pyi.merge_files(
      py_path=args.py, pyi_path=args.pyi, mode=mode, backup=backup)

  if mode == merge_pyi.Mode.OVERWRITE:
    if changed:
      print(f'Merged types to {args.py} from {args.pyi}')
    else:
      print(f'No new types for {args.py} in {args.pyi}')


if __name__ == '__main__':
  main()



from collections.abc import Generator
from typing import Any

from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver

from .extension_connection import ExtensionConnection as ExtensionConnection
from .firefox_binary import FirefoxBinary as FirefoxBinary
from .firefox_profile import FirefoxProfile as FirefoxProfile
from .options import Options as Options
from .remote_connection import FirefoxRemoteConnection as FirefoxRemoteConnection
from .service import Service as Service
from .webelement import FirefoxWebElement as FirefoxWebElement

basestring = str

class WebDriver(RemoteWebDriver):
    NATIVE_EVENTS_ALLOWED: Any
    CONTEXT_CHROME: str
    CONTEXT_CONTENT: str
    binary: Any
    profile: Any
    service: Any
    def __init__(
        self,
        firefox_profile: Any | None = ...,
        firefox_binary: Any | None = ...,
        timeout: int = ...,
        capabilities: Any | None = ...,
        proxy: Any | None = ...,
        executable_path: str = ...,
        options: Any | None = ...,
        service_log_path: str = ...,
        firefox_options: Any | None = ...,
        service_args: Any | None = ...,
        desired_capabilities: Any | None = ...,
        log_path: Any | None = ...,
        keep_alive: bool = ...,
    ) -> None: ...
    def quit(self) -> None: ...
    @property
    def firefox_profile(self): ...
    def set_context(self, context) -> None: ...
    def context(self, context) -> Generator[None, None, None]: ...
    def install_addon(self, path, temporary: Any | None = ...): ...
    def uninstall_addon(self, identifier) -> None: ...



from typing import Any

from .commands import GraphCommands as GraphCommands
from .edge import Edge as Edge
from .node import Node as Node
from .path import Path as Path

class Graph(GraphCommands):
    NAME: Any
    client: Any
    execute_command: Any
    nodes: Any
    edges: Any
    version: int
    def __init__(self, client, name=...) -> None: ...
    @property
    def name(self): ...
    def get_label(self, idx): ...
    def get_relation(self, idx): ...
    def get_property(self, idx): ...
    def add_node(self, node) -> None: ...
    def add_edge(self, edge) -> None: ...
    def call_procedure(self, procedure, *args, read_only: bool = ..., **kwagrs): ...
    def labels(self): ...
    def relationshipTypes(self): ...
    def propertyKeys(self): ...



from distutils.cmd import Command
from typing import Any, Iterable, Mapping, Text

class Distribution:
    cmdclass: dict[str, type[Command]]
    def __init__(self, attrs: Mapping[str, Any] | None = ...) -> None: ...
    def get_option_dict(self, command: str) -> dict[str, tuple[str, Text]]: ...
    def parse_config_files(self, filenames: Iterable[Text] | None = ...) -> None: ...
    def get_command_obj(self, command: str, create: bool = ...) -> Command | None: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class RichText(Serialisable):
    tagname: str
    bodyPr: Any
    properties: Any
    lstStyle: Any
    p: Any
    paragraphs: Any
    __elements__: Any
    def __init__(self, bodyPr: Any | None = ..., lstStyle: Any | None = ..., p: Any | None = ...) -> None: ...

class Text(Serialisable):
    tagname: str
    strRef: Any
    rich: Any
    __elements__: Any
    def __init__(self, strRef: Any | None = ..., rich: Any | None = ...) -> None: ...
    def to_tree(self, tagname: Any | None = ..., idx: Any | None = ..., namespace: Any | None = ...): ...



from typing import Any, Callable, Iterable, TypeVar, overload

_T = TypeVar("_T")
_S = TypeVar("_S")

@overload
def first(iterable: Iterable[_T]) -> _T | None: ...
@overload
def first(iterable: Iterable[_T], default: _S) -> _T | _S: ...
@overload
def first(iterable: Iterable[_T], default: _S, key: Callable[[_T], Any] | None) -> _T | _S: ...
@overload
def first(iterable: Iterable[_T], *, key: Callable[[_T], Any] | None) -> _T | None: ...



from typing import Any
from typing_extensions import Literal

_ImageFilter = Literal["AUTO", "FlateDecode", "DCTDecode", "JPXDecode"]

SUPPORTED_IMAGE_FILTERS: tuple[_ImageFilter, ...]

def load_image(filename): ...

# Returned dict could be typed as a TypedDict.
def get_img_info(img, image_filter: _ImageFilter = ..., dims: Any | None = ...) -> dict[str, Any]: ...



from typing import Any

from pyVmomi.vmodl import ManagedObject

def __getattr__(name: str) -> Any: ...  # incomplete

class InvalidArgument(Exception): ...

class ManagedObjectNotFound:
    obj: ManagedObject



from typing import Any

from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver

from .options import Options as Options
from .service import Service as Service

DEFAULT_TIMEOUT: int
DEFAULT_PORT: int
DEFAULT_HOST: Any
DEFAULT_LOG_LEVEL: Any
DEFAULT_SERVICE_LOG_PATH: Any

class WebDriver(RemoteWebDriver):
    port: Any
    host: Any
    iedriver: Any
    def __init__(
        self,
        executable_path: str = ...,
        capabilities: Any | None = ...,
        port=...,
        timeout=...,
        host=...,
        log_level=...,
        service_log_path=...,
        options: Any | None = ...,
        ie_options: Any | None = ...,
        desired_capabilities: Any | None = ...,
        log_file: Any | None = ...,
        keep_alive: bool = ...,
    ) -> None: ...
    def quit(self) -> None: ...
    def create_options(self): ...



from typing import Sequence

def add_uppercase_char(char_list: Sequence[tuple[str, str]]) -> Sequence[tuple[str, str]]: ...

CYRILLIC: Sequence[tuple[str, str]]
GERMAN: Sequence[tuple[str, str]]
GREEK: Sequence[tuple[str, str]]
PRE_TRANSLATIONS: Sequence[tuple[str, str]]



from abc import abstractmethod
from typing import Any

from openpyxl.descriptors.serialisable import Serialisable as Serialisable

class _DataLabelBase(Serialisable):
    numFmt: Any
    spPr: Any
    graphicalProperties: Any
    txPr: Any
    textProperties: Any
    dLblPos: Any
    position: Any
    showLegendKey: Any
    showVal: Any
    showCatName: Any
    showSerName: Any
    showPercent: Any
    showBubbleSize: Any
    showLeaderLines: Any
    separator: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        numFmt: Any | None = ...,
        spPr: Any | None = ...,
        txPr: Any | None = ...,
        dLblPos: Any | None = ...,
        showLegendKey: Any | None = ...,
        showVal: Any | None = ...,
        showCatName: Any | None = ...,
        showSerName: Any | None = ...,
        showPercent: Any | None = ...,
        showBubbleSize: Any | None = ...,
        showLeaderLines: Any | None = ...,
        separator: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...
    @property
    @abstractmethod
    def tagname(self) -> str: ...

class DataLabel(_DataLabelBase):
    tagname: str
    idx: Any
    numFmt: Any
    spPr: Any
    txPr: Any
    dLblPos: Any
    showLegendKey: Any
    showVal: Any
    showCatName: Any
    showSerName: Any
    showPercent: Any
    showBubbleSize: Any
    showLeaderLines: Any
    separator: Any
    extLst: Any
    __elements__: Any
    def __init__(self, idx: int = ..., **kw) -> None: ...

class DataLabelList(_DataLabelBase):
    tagname: str
    dLbl: Any
    delete: Any
    numFmt: Any
    spPr: Any
    txPr: Any
    dLblPos: Any
    showLegendKey: Any
    showVal: Any
    showCatName: Any
    showSerName: Any
    showPercent: Any
    showBubbleSize: Any
    showLeaderLines: Any
    separator: Any
    extLst: Any
    __elements__: Any
    def __init__(self, dLbl=..., delete: Any | None = ..., **kw) -> None: ...



import sys
from _typeshed import ReadableBuffer
from typing import Callable

from paramiko.kex_group1 import KexGroup1 as KexGroup1

if sys.version_info >= (3, 0):
    from hashlib import _Hash
else:
    from hashlib import _hash as _Hash

class KexGroup16SHA512(KexGroup1):
    name: str
    P: int
    G: int
    hash_algo: Callable[[ReadableBuffer], _Hash]



"""Implementation of named tuples."""

import dataclasses
import keyword

from typing import Any, List, Optional

from pytype import special_builtins
from pytype import utils
from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.abstract import class_mixin
from pytype.abstract import function
from pytype.overlays import classgen
from pytype.overlays import overlay_utils
from pytype.pytd import escape
from pytype.pytd import pytd
from pytype.pytd import pytd_utils
from pytype.pytd import visitors


# type alias
Param = overlay_utils.Param


# This module has classes and methods which benefit from extended docstrings,
# but whose argument list documentation is trivial and only adds clutter.
# Disable the lint check rather than convert the extended docstrings to
# one-line docstrings + comments.
# pylint: disable=g-doc-args
# pylint: disable=g-doc-return-or-yield
# pylint: disable=g-doc-exception


@dataclasses.dataclass
class Field:
  """A namedtuple field."""

  name: str
  typ: Any
  default: Any = None


@dataclasses.dataclass
class NamedTupleProperties:
  """Collection of properties used to construct a namedtuple."""

  name: str
  fields: List[Field]
  bases: List[Any]

  @classmethod
  def from_field_names(cls, name, field_names, ctx):
    """Make a NamedTupleProperties from field names with no types."""
    fields = [Field(n, ctx.convert.unsolvable, None) for n in field_names]
    return cls(name, fields, [])

  def validate_and_rename_fields(self, rename):
    """Validate and rename self.fields.

    namedtuple field names have some requirements:
    - must not be a Python keyword
    - must consist of only alphanumeric characters and "_"
    - must not start with "_" or a digit

    Basically, they're valid Python identifiers that don't start with "_" or a
    digit. Also, there can be no duplicate field names.

    If rename is true, any invalid field names are changed to "_%d". For
    example, "abc def ghi abc" becomes "abc _1 def _3" because "def" is a
    keyword and "abc" is a duplicate.

    Also validates self.name, which has the same requirements, except it can
    start with "_", and cannot be changed.
    """
    def invalid_name(field_name):
      return (not all(c.isalnum() or c == "_" for c in field_name)
              or keyword.iskeyword(field_name)
              or not field_name  # catches empty string, etc.
              or field_name[0].isdigit())

    if invalid_name(self.name):
      raise ValueError(self.name)

    seen = set()
    for idx, f in enumerate(self.fields):
      if invalid_name(f.name) or f.name.startswith("_") or f.name in seen:
        if rename:
          f.name = "_%d" % idx
        else:
          raise ValueError(f.name)
      seen.add(f.name)


@dataclasses.dataclass
class _Args:
  """Args for both collections.namedtuple and typing.NamedTuple."""

  name: str
  field_names: List[str]
  field_types: Optional[List[Any]] = None
  defaults: Optional[List[Any]] = None
  rename: bool = False


class _ArgsError(Exception):
  pass


class _FieldMatchError(Exception):
  """Errors when postprocessing field args, to be converted to WrongArgTypes."""

  def __init__(self, param):
    super().__init__()
    self.param = param


class NamedTupleBuilderBase(abstract.PyTDFunction):
  """Base class that handles namedtuple function args processing.

  Performs the same argument checking as collections.namedtuple, e.g. making
  sure field names don't start with _ or digits, making sure no keywords are
  used for the typename or field names, and so on.

  If incorrect arguments are passed, a subclass of function.FailedFunctionCall
  will be raised. Other cases (e.g. if any of the arguments have multiple
  bindings) raise _ArgsError(). This also occurs if an argument value is
  invalid, e.g. a keyword is used as a field name and rename is False.
  Subclasses catch _ArgsError() and return Any in place of a named tuple class.
  """

  # Args are processed in several stages:
  #   raw_args: The incoming function.Args passed to the constructor.
  #   callargs: A dict of {key: unwrapped arg value}
  #   _Args: Postprocessed arg values struct, unifies collections.namedtuple and
  #     the typing.NamedTuple functional constructor
  #   NamedTupleProperties: The final list of properties passed to the class
  #     builder. Unifies the functional constructors with the subclass
  #     constructor (which has class annotations rather than function args).

  def extract_args(self, node, callargs):
    """Extract callargs into an _Args object.

    Subclasses should implement extract_args for their specific args.
    """
    raise NotImplementedError()

  def process_args(self, node, raw_args):
    """Convert namedtuple call args into a NamedTupleProperties.

    Returns both the NamedTupleProperties and an _Args struct in case the caller
    wants to do any more args processing before calling the class builder.
    """

    # Check the args against the pytd signature
    self.match_args(node, raw_args)

    # namedtuple only has one signature
    sig, = self.signatures

    try:
      callargs = {name: abstract_utils.get_atomic_python_constant(var)
                  for name, var, _ in sig.signature.iter_args(raw_args)}
      args = self.extract_args(node, callargs)
    except abstract_utils.ConversionError:
      raise _ArgsError()  # pylint: disable=raise-missing-from
    except _FieldMatchError as e:
      raise function.WrongArgTypes(sig.signature, raw_args, self.ctx, e.param)

    props = NamedTupleProperties.from_field_names(
        args.name, args.field_names, self.ctx)
    try:
      props.validate_and_rename_fields(args.rename)
    except ValueError as e:
      self.ctx.errorlog.invalid_namedtuple_arg(
          self.ctx.vm.frames, utils.message(e))
      raise _ArgsError()  # pylint: disable=raise-missing-from

    if args.defaults:
      for f, d in zip(props.fields, args.defaults):
        f.default = self.ctx.new_unsolvable(node) if d else None

    return args, props


class NamedTupleBuilder(NamedTupleBuilderBase):
  """Factory for creating collections.namedtuple classes."""

  collections_ast: pytd.TypeDeclUnit

  @classmethod
  def make(cls, name, ctx, pyval=None):
    # Loading the ast should be memoized after the import in CollectionsOverlay
    collections_ast = ctx.loader.import_name("collections")
    # Subclasses of NamedTupleBuilder need a different pyval.
    if not pyval:
      pyval = collections_ast.Lookup("collections.namedtuple")
    self = super().make(name, ctx, "collections", pyval=pyval)  # pytype: disable=wrong-arg-types
    self.collections_ast = collections_ast
    return self

  def extract_args(self, node, callargs):
    """Extracts the typename, field_names and rename arguments.

    collections.namedtuple takes a 'verbose' argument too but we don't care
    about that.

    The 'defaults' arg is postprocessed from a sequence of defaults to a
    sequence of bools describing whether each field has a default (e.g., for
    collections.namedtuple('X', field_names=['a', 'b'], defaults=[0]) this
    method will return [False, True] for defaults to indicate that 'a' does not
    have a default while 'b' does).
    """
    name = callargs["typename"]

    # namedtuple fields can be given as a single string, e.g. "a, b, c" or as a
    # list [Variable('a'), Variable('b'), Variable('c')].
    # We just want a list of strings.
    fields = callargs["field_names"]
    if isinstance(fields, (bytes, str)):
      fields = utils.native_str(fields)
      field_names = fields.replace(",", " ").split()
    else:
      field_names = [abstract_utils.get_atomic_python_constant(f)
                     for f in fields]
      field_names = [utils.native_str(f) for f in field_names]

    if "defaults" in callargs:
      default_vars = callargs["defaults"]
      num_defaults = len(default_vars)
      defaults = [False] * (len(fields) - num_defaults) + [True] * num_defaults
    else:
      defaults = [False] * len(fields)

    # rename will take any problematic field names and give them a new name.
    rename = callargs.get("rename", False)

    return _Args(
        name=name, field_names=field_names, defaults=defaults, rename=rename)

  def call(self, node, _, args):
    """Creates a namedtuple class definition."""
    # If we can't extract the arguments, we take the easy way out and return Any
    try:
      _, props = self.process_args(node, args)
    except _ArgsError:
      return node, self.ctx.new_unsolvable(node)

    node, cls_var = _build_namedtuple(props, node, self.ctx)
    return node, cls_var


class NamedTupleFuncBuilder(NamedTupleBuilderBase):
  """Factory for creating typing.NamedTuples via the function constructor."""

  _fields_param: function.BadParam

  @classmethod
  def make(cls, ctx):
    typing_ast = ctx.loader.import_name("typing")
    # typing.pytd contains a NamedTuple class def and a _NamedTuple func def.
    # Replace the name of the returned function so that error messages will
    # correctly display "typing.NamedTuple".
    pyval = typing_ast.Lookup("typing._NamedTuple")
    pyval = pyval.Replace(name="typing.NamedTuple")
    self = super().make("NamedTuple", ctx, "typing", pyval)
    # NamedTuple's fields arg has type Sequence[Sequence[Union[str, type]]],
    # which doesn't provide precise enough type-checking, so we have to do
    # some of our own in _getargs. _NamedTupleFields is an alias to
    # List[Tuple[str, type]], which gives a more understandable error message.
    fields_pyval = typing_ast.Lookup("typing._NamedTupleFields").type
    fields_type = ctx.convert.constant_to_value(fields_pyval, {}, ctx.root_node)
    # pylint: disable=protected-access
    self._fields_param = function.BadParam(name="fields", expected=fields_type)
    return self

  def _is_str_instance(self, val):
    return (isinstance(val, abstract.Instance) and
            val.full_name in ("builtins.str", "builtins.unicode"))

  def extract_args(self, node, callargs):
    """Extracts the typename and fields arguments.

    fields is postprocessed into field_names and field_types.

    typing.NamedTuple doesn't support rename, it will default to False
    """

    cls_name = callargs["typename"]

    fields = callargs["fields"]
    if isinstance(fields, str):
      # Since str matches Sequence, we have to manually check for it.
      raise _FieldMatchError(self._fields_param)
    # The fields is a list of tuples, so we need to deeply unwrap them.
    fields = [abstract_utils.get_atomic_python_constant(t) for t in fields]
    # We need the actual string for the field names and the BaseValue
    # for the field types.
    names = []
    types = []
    for field in fields:
      if isinstance(field, str):
        # Since str matches Sequence, we have to manually check for it.
        raise _FieldMatchError(self._fields_param)
      if (len(field) != 2 or
          any(not self._is_str_instance(v) for v in field[0].data)):
        # Note that we don't need to check field[1] because both 'str'
        # (forward reference) and 'type' are valid for it.
        raise _FieldMatchError(self._fields_param)
      name, typ = field
      name_py_constant = abstract_utils.get_atomic_python_constant(name)
      names.append(name_py_constant)
      annot = self.ctx.annotation_utils.extract_annotation(
          node,
          typ,
          name_py_constant,
          self.ctx.vm.simple_stack(),
          allowed_type_params=set())
      types.append(annot)

    return _Args(name=cls_name, field_names=names, field_types=types)

  def call(self, node, _, args):
    try:
      args, props = self.process_args(node, args)
    except _ArgsError:
      return node, self.ctx.new_unsolvable(node)

    # fill in field types from annotations
    annots = self.ctx.annotation_utils.convert_annotations_list(
        node, zip(args.field_names, args.field_types))
    for f in props.fields:
      f.typ = annots.get(f.name, self.ctx.convert.unsolvable)

    node, cls_var = _build_namedtuple(props, node, self.ctx)
    return node, cls_var


class NamedTupleClassBuilder(abstract.PyTDClass):
  """Factory for creating typing.NamedTuples by subclassing NamedTuple."""

  # attributes prohibited to set in NamedTuple class syntax
  _prohibited = ("__new__", "__init__", "__slots__", "__getnewargs__",
                 "_fields", "_field_defaults", "_field_types",
                 "_make", "_replace", "_asdict", "_source")

  def __init__(self, ctx):
    typing_ast = ctx.loader.import_name("typing")
    pyval = typing_ast.Lookup("typing.NamedTuple")
    super().__init__("NamedTuple", pyval, ctx)
    # Prior to python 3.6, NamedTuple is a function. Although NamedTuple is a
    # class in python 3.6+, we can still use it like a function. Hold the
    # an instance of 'NamedTupleFuncBuilder' so that we can reuse the
    # old implementation to implement the NamedTuple in python 3.6+
    self.namedtuple = NamedTupleFuncBuilder.make(ctx)

  def call(self, node, _, args):
    posargs = args.posargs
    if isinstance(args.namedargs, dict):
      namedargs = args.namedargs
    else:
      namedargs = self.ctx.convert.value_to_constant(args.namedargs, dict)
    if namedargs and len(posargs) == 1:
      namedargs = [
          self.ctx.convert.build_tuple(
              node, (self.ctx.convert.build_string(node, k), v))
          for k, v in namedargs.items()
      ]
      namedargs = abstract.List(namedargs, self.ctx).to_variable(node)
      posargs += (namedargs,)
      args = function.Args(posargs)
    elif namedargs:
      errmsg = ("Either list of fields or keywords can be provided to "
                "NamedTuple, not both")
      self.ctx.errorlog.invalid_namedtuple_arg(
          self.ctx.vm.frames, err_msg=errmsg)
    return self.namedtuple.call(node, None, args)

  def make_class(self, node, bases, f_locals):
    # If BuildClass.call() hits max depth, f_locals will be [unsolvable]
    # Since we don't support defining NamedTuple subclasses in a nested scope
    # anyway, we can just return unsolvable here to prevent a crash, and let the
    # invalid namedtuple error get raised later.
    if isinstance(f_locals.data[0], abstract.Unsolvable):
      return node, self.ctx.new_unsolvable(node)

    f_locals = abstract_utils.get_atomic_python_constant(f_locals)

    name = f_locals["__qualname__"]
    name = abstract_utils.get_atomic_python_constant(name)
    if "." in name:
      name = name.rsplit(".", 1)[-1]

    # Construct a NamedTupleProperties to pass to the builder function.
    cls_locals = classgen.get_class_locals(
        name,
        allow_methods=True,
        ordering=classgen.Ordering.FIRST_ANNOTATE,
        ctx=self.ctx)
    props = NamedTupleProperties(name=name, fields=[], bases=bases)
    stack = tuple(self.ctx.vm.frames)
    for k, local in cls_locals.items():
      assert local.typ
      t = self.ctx.annotation_utils.extract_annotation(
          node, local.typ, k, stack)
      props.fields.append(Field(name=k, typ=t, default=f_locals.get(k)))

    # typing.NamedTuple doesn't support rename; invalid fields are an error.
    try:
      props.validate_and_rename_fields(rename=False)
    except ValueError as e:
      self.ctx.errorlog.invalid_namedtuple_arg(
          self.ctx.vm.frames, utils.message(e))
      return node, self.ctx.new_unsolvable(node)

    node, cls_var = _build_namedtuple(props, node, self.ctx)
    cls_val = abstract_utils.get_atomic_value(cls_var)

    if not isinstance(cls_val, abstract.Unsolvable):

      # set the attribute without overriding special namedtuple attributes
      node, fields = self.ctx.attribute_handler.get_attribute(
          node, cls_val, "_fields")
      fields = abstract_utils.get_atomic_python_constant(fields, tuple)
      fields = [abstract_utils.get_atomic_python_constant(field, str)
                for field in fields]
      for key in f_locals:
        if key in self._prohibited:
          self.ctx.errorlog.not_writable(self.ctx.vm.frames, cls_val, key)
        if key not in abstract_utils.CLASS_LEVEL_IGNORE and  key not in fields:
          node = self.ctx.attribute_handler.set_attribute(
              node, cls_val, key, f_locals[key])

    return node, cls_var

  def make_class_from_pyi(self, cls_name, pytd_cls):
    """Make a NamedTupleClass from a pyi class."""
    # NOTE: Returns the abstract class, not a variable.
    ctx = self.ctx
    fields = []
    classvars = []
    for c in pytd_cls.constants:
      ct = c.type.base_type if isinstance(c.type, pytd.Annotated) else c.type
      if isinstance(ct, pytd.GenericType):
        cv = pytd_utils.UnpackGeneric(ct, "typing.ClassVar")
      elif ct.name == "typing.ClassVar":
        cv = (pytd.AnythingType(),)
      else:
        cv = None
      if cv is not None:
        typ, = cv
        classvars.append((c.name, typ))
      else:
        # The field types may refer back to the class being built.
        with ctx.allow_recursive_convert():
          if c.value:
            default = ctx.new_unsolvable(ctx.root_node)
          else:
            default = None
          fields.append(
              Field(c.name, ctx.convert.constant_to_value(ct), default))

    bases = []
    for x in pytd_cls.bases:
      b = ctx.convert.constant_to_var(x)
      if pytd_cls.template:
        # We need to set the generated class's _template directly here, since we
        # bypass the normal mechanism to copy a template from a pytd class to a
        # PyTDClass (see abstract._base._compute_template())
        tmpl = [ctx.convert.constant_to_value(t.type_param)
                for t in pytd_cls.template]
        b.data[0]._template = tmpl  # pylint: disable=protected-access
      bases.append(b)
    props = NamedTupleProperties(cls_name, fields, bases)
    node = ctx.root_node
    node, cls_var = _build_namedtuple(props, node, ctx)
    cls = abstract_utils.get_atomic_value(cls_var)

    # Adjust the class after construction

    # Since we might have classvars or methods referring to the class itself,
    # we need to insert the just-constructed class into the convert cache.
    ctx.convert._convert_cache[("constant", pytd_cls, type(pytd_cls))] = cls  # pylint: disable=protected-access

    # Set the official class name.
    cls.update_official_name(cls.name)

    # Convert classvars, which might be recursively typed
    with ctx.allow_recursive_convert():
      classvars = [(name, ctx.convert.constant_to_value(typ))
                   for name, typ in classvars]

    # Add fields and classvars to the annotation dictionary
    locals_ = {f.name: abstract_utils.Local(node, None, f.typ, None, ctx)
               for f in fields}
    locals_.update({name: abstract_utils.Local(node, None, typ, None, ctx)
                    for name, typ in classvars})
    annots = abstract.AnnotationsDict(locals_, ctx).to_variable(node)
    cls.members["__annotations__"] = annots

    # Add classvars to members
    for name, typ in classvars:
      _, var = ctx.vm.init_class(ctx.root_node, typ)
      cls.members[name] = var

    # Add any methods defined in the class, potentially overwriting the
    # generated ones
    for m in pytd_cls.methods:
      # We need to allow recursive conversion because overriding __new__ adds a
      # typevar that refers back to the class, and reloading the pyi file
      # triggers a recursion error in convert.
      with ctx.allow_recursive_convert():
        # Create a SimpleFunction from the pytd method; storing a PyTDFunction
        # as an InterpreterClass method does not behave correctly (see
        # AttributeHandler._lookup_from_mro)
        sig = function.Signature.from_pytd(ctx, m.name, m.signatures[0])
        meth = abstract.SimpleFunction(sig, ctx)
        m_var = meth.to_variable(ctx.root_node)
        # Handle classmethods and staticmethods
        # TODO(mdemello): We really need a better way to do this.
        # Move to SimpleFunction.from_pyi if we add it to one more overlay.
        args = function.Args(posargs=(m_var,))
        if m.kind == pytd.MethodKind.CLASSMETHOD:
          _, m_var = special_builtins.ClassMethod(ctx).call(node, meth, args)
        elif m.kind == pytd.MethodKind.STATICMETHOD:
          _, m_var = special_builtins.StaticMethod(ctx).call(node, meth, args)
        cls.members[m.name] = m_var

    return cls


class _DictBuilder:
  """Construct dict abstract classes for namedtuple members."""

  def __init__(self, ctx):
    self.ctx = ctx
    self.dict_cls = ctx.convert.name_to_value("builtins.dict")

  def make(self, typ):
    # Normally, we would use abstract_utils.K and abstract_utils.V, but
    # collections.pyi doesn't conform to that standard.
    return abstract.ParameterizedClass(self.dict_cls, {
        "K": self.ctx.convert.str_type,
        "V": typ
    }, self.ctx)


class NamedTupleClass(abstract.InterpreterClass):
  """Named tuple classes."""

  def __init__(self, *args):
    super().__init__(*args)
    # Store the original properties, to output to pyi files.
    self.props = None
    self.generated_members = None

  def instantiate(self, node, container=None):
    # For all generic typevars T, add a type parameter alias between cls.T and
    # path.to.module.cls.T
    inst = super().instantiate(node, container)
    for ival in inst.data:
      cls = ival.cls
      long = cls.full_name
      for t in cls.template:
        short = t.module
        param = t.name
        ival.instance_type_parameters.add_alias(
            f"{short}.{param}", f"{long}.{param}", lambda x, y: x or y)
    return inst


def _build_namedtuple(props, node, ctx):
  """Build an InterpreterClass representing the namedtuple."""

  # TODO(mdemello): Fix this to support late types.
  if props.fields and props.fields[0].typ:
    field_types_union = abstract.Union([f.typ for f in props.fields], ctx)
  else:
    field_types_union = ctx.convert.none_type

  members = {f.name: f.typ.instantiate(node) for f in props.fields}

  # NOTE: We add the full list of private methods to all namedtuples.
  # Technically collections.namedtuple has a smaller set.
  # collections.namedtuple has: __dict__, __slots__ and _fields.
  # typing.NamedTuple adds: _field_types, __annotations__ and _field_defaults.
  # __slots__ and _fields are tuples containing the names of the fields.
  slots = tuple(ctx.convert.build_string(node, f.name) for f in props.fields)
  members["__slots__"] = ctx.convert.build_tuple(node, slots)
  members["_fields"] = ctx.convert.build_tuple(node, slots)

  odict = _DictBuilder(ctx)
  # __dict__ and _field_defaults are both dicts of
  # { field_name: field_type_instance }
  # The field types may refer back to the class being built.
  with ctx.allow_recursive_convert():
    field_dict_cls = odict.make(field_types_union)
  members["__dict__"] = field_dict_cls.instantiate(node)
  members["_field_defaults"] = field_dict_cls.instantiate(node)

  # _field_types and __annotations__ are both dicts of
  # { field_name: field_type }
  # Note that ctx.make_class will take care of adding the __annotations__
  # member.
  field_types_cls = odict.make(ctx.convert.type_type)
  members["_field_types"] = field_types_cls.instantiate(node)

  # __new__
  # We set the bound on this TypeParameter later. This gives __new__ the
  # signature: def __new__(cls: Type[_Tname], ...) -> _Tname, i.e. the same
  # signature that visitor.CreateTypeParametersForSignatures would create.
  # This allows subclasses of the NamedTuple to get the correct type from
  # their constructors.
  # The TypeParameter name is built from the class name and field names to avoid
  # name clashes with other namedtuples.
  cls_type_param_name = (
      visitors.CreateTypeParametersForSignatures.PREFIX +
      escape.pack_namedtuple(props.name, [f.name for f in props.fields]))
  cls_type_param = abstract.TypeParameter(cls_type_param_name, ctx, bound=None)
  cls_type = abstract.ParameterizedClass(ctx.convert.type_type,
                                         {abstract_utils.T: cls_type_param},
                                         ctx)

  params = [Param(f.name, f.typ) for f in props.fields]
  # The parameter types may refer back to the class being built.
  with ctx.allow_recursive_convert():
    members["__new__"] = overlay_utils.make_method(
        ctx,
        node,
        name="__new__",
        self_param=Param("cls", cls_type),
        params=params,
        return_type=cls_type_param,
    )

  # __init__
  members["__init__"] = overlay_utils.make_method(
      ctx,
      node,
      name="__init__",
      varargs=Param("args"),
      kwargs=Param("kwargs"))

  heterogeneous_tuple_type_params = dict(enumerate(f.typ for f in props.fields))
  heterogeneous_tuple_type_params[abstract_utils.T] = field_types_union
  # Representation of the to-be-created NamedTuple as a typing.Tuple.
  heterogeneous_tuple_type = abstract.TupleClass(
      ctx.convert.tuple_type, heterogeneous_tuple_type_params, ctx)

  # _make
  # _make is a classmethod, so it needs to be wrapped by
  # special_builtins.ClassMethodInstance.
  # Like __new__, it uses the _Tname TypeVar.
  sized_cls = ctx.convert.name_to_value("typing.Sized")
  iterable_type = abstract.ParameterizedClass(
      ctx.convert.name_to_value("typing.Iterable"),
      {abstract_utils.T: field_types_union}, ctx)
  cls_type = abstract.ParameterizedClass(ctx.convert.type_type,
                                         {abstract_utils.T: cls_type_param},
                                         ctx)
  len_type = abstract.CallableClass(
      ctx.convert.name_to_value("typing.Callable"), {
          0: sized_cls,
          abstract_utils.ARGS: sized_cls,
          abstract_utils.RET: ctx.convert.int_type
      }, ctx)
  params = [
      Param("iterable", iterable_type),
      Param("new").unsolvable(ctx, node),
      Param("len", len_type).unsolvable(ctx, node)
  ]
  make = overlay_utils.make_method(
      ctx,
      node,
      name="_make",
      params=params,
      self_param=Param("cls", cls_type),
      return_type=cls_type_param)
  make_args = function.Args(posargs=(make,))
  _, members["_make"] = ctx.special_builtins["classmethod"].call(
      node, None, make_args)

  # _replace
  # Like __new__, it uses the _Tname TypeVar. We have to annotate the `self`
  # param to make sure the TypeVar is substituted correctly.
  members["_replace"] = overlay_utils.make_method(
      ctx,
      node,
      name="_replace",
      self_param=Param("self", cls_type_param),
      return_type=cls_type_param,
      kwargs=Param("kwds", field_types_union))

  # __getnewargs__
  members["__getnewargs__"] = overlay_utils.make_method(
      ctx,
      node,
      name="__getnewargs__",
      return_type=heterogeneous_tuple_type)

  # __getstate__
  members["__getstate__"] = overlay_utils.make_method(
      ctx, node, name="__getstate__")

  # _asdict
  members["_asdict"] = overlay_utils.make_method(
      ctx, node, name="_asdict", return_type=field_dict_cls)

  # Finally, make the class.
  cls_dict = abstract.Dict(ctx)
  cls_dict.update(node, members)

  if ctx.options.strict_namedtuple_checks:
    # Enforces type checking like Tuple[...]
    superclass_of_new_type = heterogeneous_tuple_type.to_variable(node)
  else:
    superclass_of_new_type = ctx.convert.tuple_type.to_variable(node)
  if props.bases:
    final_bases = []
    for base in props.bases:
      if any(b.full_name == "typing.NamedTuple" for b in base.data):
        final_bases.append(superclass_of_new_type)
      else:
        final_bases.append(base)
  else:
    final_bases = [superclass_of_new_type]
    # This NamedTuple is being created via a function call. We manually
    # construct an annotated_locals entry for it so that __annotations__ is
    # initialized properly for the generated class.
    ctx.vm.annotated_locals[props.name] = {
        f.name: abstract_utils.Local(node, None, f.typ, None, ctx)
        for f in props.fields
    }

  cls_props = class_mixin.ClassBuilderProperties(
      name_var=ctx.convert.build_string(node, props.name),
      bases=final_bases,
      class_dict_var=cls_dict.to_variable(node),
      class_type=NamedTupleClass,)
  node, cls_var = ctx.make_class(node, cls_props)
  cls = cls_var.data[0]
  # Now that the class has been made, we can complete the TypeParameter used
  # by __new__, _make and _replace.
  cls_type_param.bound = cls

  # set __new__.__defaults__
  defaults = [f.default for f in props.fields if f.default is not None]
  defaults = ctx.convert.build_tuple(node, defaults)
  node, new_attr = ctx.attribute_handler.get_attribute(
      node, cls, "__new__")
  new_attr = abstract_utils.get_atomic_value(new_attr)
  node = ctx.attribute_handler.set_attribute(
      node, new_attr, "__defaults__", defaults)

  # Store the original properties
  cls.props = props
  cls.generated_members = (
      set(members.keys()) - {x.name for x in props.fields})

  ctx.vm.trace_classdef(cls_var)
  return node, cls_var
